"""
A class to handle the collection of run statistics and information 
from fastq files and upload to minotour.
"""
import logging
import os
import sys
from collections import defaultdict

from minFQ.endpoints import EndPoint
from minFQ.minotourapi import MinotourAPI

import toml as toml_manager

log = logging.getLogger(__name__)


def get_flowcell_name_from_desc(description_dict, user_run_name):
    """
    Get the flowcell name from the description
    Parameters
    ----------
    description_dict: dict
        A parsed dictionary created from the description from the fastq record
    user_run_name: str
        The user run name that we have been given on the command line

    Returns
    -------
    flowcell_name: str
        The flowcell name that we are gonna be using
    """
    if "flow_cell_id" in description_dict:
        flowcell_name = description_dict["flow_cell_id"]
    elif "sample_id" in description_dict:
        flowcell_name = description_dict["sample_id"]
    elif "sampleid" in description_dict:
        flowcell_name = description_dict["sampleid"]
    else:
        flowcell_name = user_run_name
    return flowcell_name


def get_unique_name(description_dict, user_run_name):
    """

    Parameters
    ----------
    description_dict: dict
        A parsed dictionary created from the description from the fastq record
    user_run_name: str
        The user run name that we have been given on the command line
    Returns
    -------

    """
    flowcell_name = user_run_name
    if "flow_cell_id" in description_dict and "sample_id" in description_dict:
        flowcell_name = "{}_{}".format(
            description_dict["flow_cell_id"], description_dict["sample_id"]
        )
    elif "flow_cell_id" in description_dict and "sampleid" in description_dict:
        flowcell_name = "{}_{}".format(
            description_dict["flow_cell_id"], description_dict["sampleid"]
        )
    return flowcell_name


def _prepare_toml(toml_dict):
    """
    Prepares the dictionary of the toml, places the channel number as key and conditon name as value
    Returns
    -------
    dict
    """
    _d = {}
    # Reverse the dict so we can lookup name by channel
    for key in toml_dict["conditions"].keys():
        channels = toml_dict["conditions"][key]["channels"]
        name = toml_dict["conditions"][key]["name"]
        _d.update({channel: name for channel in channels})
    return _d


class RunDataTracker:
    def __init__(self, args, header, sequencing_statistics):
        log.debug("Initialising Runcollection")
        self.args = args
        self.header = header
        self.read_names = []
        self.read_count = 0
        self.base_count = 0
        self.tracked_base_count = 0
        self.read_type_dict = {}
        if self.args.skip_sequence:
            self.batchsize = 5000
        else:
            self.batchsize = 4000
        self.run = None
        self.flowcell = None
        self.barcode_dict = {}
        self.read_list = []
        self.file_monitor = {}
        self.fastq_file_id = None
        self.minotour_api = MinotourAPI(
            self.args.host_name, self.args.port_number, self.header
        )
        self.get_readtype_list()
        self.unblocked_dict = {}
        self.unblocked_file = None
        self.unblocked_line_start = 0
        self.run_folder = None
        self.toml = None
        self.sequencing_statistics = sequencing_statistics

    def add_toml_file(self, toml_file):
        self.toml = toml_file

    def add_run_folder(self, folder):
        """
        Adds the folder directory for this run.
        Parameters
        ----------
        folder: str
            Path to folder that run is being written into

        Returns
        -------

        """
        self.run_folder = folder
        self.check_for_read_until_files()

    def add_unblocked_reads_file(self, unblocked_file_path):
        self.unblocked_file = unblocked_file_path

    def create_jobs(self, flowcell, run, args):
        """
        Create jobs when runCollection is initialised
        Returns
        -------

        """
        payload = {
            "flowcell": flowcell["id"],
            "job_type": args.job,
            "reference": args.reference,
            "target_set": args.targets,
            "primer_scheme": args.primer_scheme,
            "cli": True,
            "api_key": args.api_key,
            # "run": run["id"]
        }
        self.minotour_api.post(
            EndPoint.JOBS,
            json=payload
        )

    def get_readtype_list(self):
        """
        Get a list of possible read types from minotour server
        Returns
        -------
        None
        """
        # fix me manual get
        read_type_list = self.minotour_api.get_json(EndPoint.READ_TYPES)
        self.read_type_dict = {read_type["name"]: read_type["id"] for read_type in read_type_list}

    def get_readnames_by_run(self, fastq_file_id):
        """
        Get read ids for the last uncomplete file we have in a run
        Parameters
        ----------
        fastq_file_id: int
            The id of the fastq file

        Returns
        -------

        """
        if fastq_file_id != self.fastq_file_id:
            self.fastq_file_id = fastq_file_id
            read_name_list = self.minotour_api.get_json(EndPoint.READ_NAMES, base_id=fastq_file_id)
            number_pages = read_name_list["number_pages"]
            log.debug("Fetching reads to check if we've uploaded these before.")
            log.debug("Wiping previous reads seen.")
            self.read_names = []
            for page in range(number_pages):
                self.sequencing_statistics.fastq_message = "Fetching {} of {} pages.".format(
                    page, number_pages
                )
                page_query = "?page={}".format(page)
                partial_file_content = self.minotour_api.get_json(EndPoint.READ_NAMES, base_id=fastq_file_id, params=page_query)
                # We have to recover the data component and loop through that to get the read names.
                for read in partial_file_content["data"]:
                    self.read_names.append(read)
            log.debug(
                "{} reads already processed and included into readnames list for run {}".format(
                    len(self.read_names), self.run["id"]
                )
            )

    def add_run(self, description_dict, args):
        """
        Add run to the minoTOur server
        Parameters
        ----------
        description_dict: dict
            Dictionary of a fastq read description
        args: argparse.NameSpace
            The namespace of the arguments on the CLI

        Returns
        -------

        """
        self.sequencing_statistics.fastq_message = "Adding run."
        run_id = description_dict["runid"]
        run = self.minotour_api.get_json(EndPoint.RUNS, base_id=run_id, params="search_criteria=runid")
        flowcell_name = get_flowcell_name_from_desc(description_dict, self.args.run_name_manual)
        if args.force_unique:
            flowcell_name = get_unique_name(description_dict, self.args.run_name_manual)
        if not flowcell_name:
            self.sequencing_statistics.errored = True
            self.sequencing_statistics.error_message = "Flowcell name is required. This may be old FASTQ data, please provide a name with -n."
            sys.exit("Flowcell name is required. This may be old FASTQ data, please provide a name with -n.")
        flowcell = self.minotour_api.get_json(EndPoint.FLOWCELL, base_id=flowcell_name,
                                              params="search_criteria=name")["data"]
        if not flowcell:
            flowcell = self.minotour_api.post(EndPoint.FLOWCELL, no_id=True, json={"name": flowcell_name})
        self.flowcell = flowcell

        if not run:
            # get or create a flowcell
            log.info("Looking for flowcell {}".format(flowcell_name))
            is_barcoded = True if "barcode" in description_dict.keys() else False
            has_fastq = False if self.args.skip_sequence else True
            key = "sample_id" if "sample_id" in description_dict else "sampleid"
            run_name = description_dict[key] if key in description_dict else self.args.run_name_manual
            payload = {
                "name": run_name,
                "sample_name": run_name,
                "runid": run_id,
                # TODO is this okay??
                "is_barcoded": is_barcoded,
                "has_fastq": has_fastq,
                "flowcell": flowcell["url"],
            }
            created_run = self.minotour_api.post(
                EndPoint.RUNS, no_id=True, json=payload
            )
            if not created_run:
                log.critical("There is a problem creating run")
                sys.exit()
            else:
                log.debug("run created")
            run = created_run

        if args.job:
            self.create_jobs(flowcell, run, args)
        if not self.run:
            self.run = run
        for item in run["barcodes"]:
            self.barcode_dict.update({item["name"]: item["id"]})

    def commit_reads(self):
        """
        Call create reads - which posts a batch of reads to the server.
        Returns
        -------
        None
        """
        if self.read_list:
            self.minotour_api.post(EndPoint.READS, json=self.read_list)
        for read in self.read_list:
            self.sequencing_statistics.fastq_info[read["run_id"]]["reads_uploaded"] += 1
        self.sequencing_statistics.reads_uploaded += len(self.read_list)
        # Refresh the read list
        self.read_list = []

    def update_read_type(self, read_id, type):
        """
        Update read type for complement read from default of template if we have a 1d^2 run
        Parameters
        ----------
        read_id: str
            UUid read id for this complement read
        type: str
            Complement read type (default "C")

        Returns
        -------

        """
        # payload = {"type": type}
        # # TODO check request usage
        # updateread = requests.patch(
        #     self.args.full_host
        #     + "api/v1/runs/"
        #     + str(self.runid)
        #     + "/reads/"
        #     + str(read_id)
        #     + "/",
        #     headers=self.header,
        #     json=payload,
        # )
        for _ in range(3):
            log.error("HELP - 1D^2 not supported, contact @mattloose on twitter")

    def check_1d2(self, read_id):
        """
        Check if the read is 1d&2 by checking the length of the read_id uuid
        Parameters
        ----------
        read_id: str
            The read id to update
        Returns
        -------
        bool
            True if read is 1d^2
        """
        if len(read_id) > 64:
            return True
        return False

    def add_read(self, fastq_read_payload):
        """
        Add a read to to the readnames list that we will commit to the minoTour server
        Create any non present barcodes on the minoTour server
        Parameters
        ----------
        fastq_read_payload: dict
            The fastq read dictionary that will be added to the upload list

        Returns
        -------
        None
        """
        barcode_name = fastq_read_payload["barcode_name"]
        rejected_barcode_name = fastq_read_payload["rejected_barcode_name"]
        run_id = fastq_read_payload["run_id"]
        read_id = fastq_read_payload["read_id"]
        fastq_read_payload["run"] = self.run["id"]
        # Here we are going to create the sequenced/unblocked barcodes and read barcode
        for barcode_name in [barcode_name, rejected_barcode_name]:
            if barcode_name not in self.barcode_dict:
                barcode = self.minotour_api.post(EndPoint.BARCODE, json={"name": barcode_name, "run": self.run["url"]})
                if barcode:
                    self.barcode_dict.update({barcode["name"]: barcode["id"]})
                else:
                    log.critical("Problem finding barcodes.")
                    sys.exit()

        fastq_read_payload["barcode"] = self.barcode_dict[
            fastq_read_payload["barcode_name"]
        ]
        fastq_read_payload["rejected_barcode"] = self.barcode_dict[
            fastq_read_payload["rejected_barcode_name"]
        ]
        if read_id not in self.read_names:
            if self.check_1d2(read_id):
                firstread, secondread = (
                    read_id[: len(read_id) // 2],
                    read_id[len(read_id) // 2 :],
                )
                self.update_read_type(secondread, self.read_type_dict["C"])
                fastq_read_payload["type"] = self.read_type_dict["1D^2"]
            else:
                fastq_read_payload["type"] = self.read_type_dict["T"]
            self.read_names.append(read_id)
            self.base_count += fastq_read_payload["sequence_length"]
            self.tracked_base_count += fastq_read_payload["sequence_length"]
            self.read_list.append(fastq_read_payload)
            log.debug(
                "Checking read_list size {} - {}".format(
                    len(self.read_list), self.batchsize
                )
            )
            # if we have a number of reads more than our batch size
            if len(self.read_list) >= self.batchsize:
                # aiming for yield of 100 Mb in number of reads
                log.debug("Commit reads")
                if not self.args.skip_sequence:
                    self.batchsize = int(
                        1000000 / (int(self.base_count) / self.read_count)
                    )
                self.commit_reads()
            elif self.tracked_base_count >= 1000000:
                self.commit_reads()
                self.tracked_base_count = 0
            self.read_count += 1
        else:
            print("Skipping read")
            self.args.reads_skipped += 1

    def check_for_read_until_files(self):
        """
        Check run folder for a channels.toml and an unblocked read_ids file when a run folder is set in the run collection
        Returns
        -------

        """
        log.debug("Checking for Toml")
        if self.toml is None or self.unblocked_file:
            for path, dirs, files in os.walk(self.run_folder):
                for file in files:
                    if file.endswith("channels.toml"):
                        if self.toml is None:
                            toml_dict = toml_manager.load(os.path.join(path, file))
                            self.toml = _prepare_toml(toml_dict)
                    if file == "unblocked_read_ids.txt":
                        if self.unblocked_file is None:
                            self.add_unblocked_reads_file(os.path.join(path, file))



import logging
import os
import time
import datetime
from collections import Counter
from pprint import pformat

from google.protobuf.json_format import MessageToJson
import pytz
from grpc import RpcError
from urllib.parse import urlparse

from minknow_api.protocol_pb2 import ProtocolRunUserInfo

from minFQ.endpoints import EndPoint

log = logging.getLogger(__name__)


def check_warnings(device_type, minknow_version):
    """
    Check to see if there any compatability warnings for minknow monitorng
    Parameters
    ----------
    device_type: str
        The device name for the sequencer connected to minKNOW. May be one of PROMETHION, GRIDION, MINION
    minknow_version: str
        The version of minknow that we are using.

    Returns
    -------

    """
    if device_type == "PROMETHION":
        log.warning("This version of minFQ may not be compatible with PromethION.")
    if minknow_version.startswith("3.3"):
        log.warning(
            "This version of minFQ may not be compatible with the MinKNOW version you are running."
        )
        log.warning(
            "As a consequence, live monitoring MAY NOT WORK. We recommend using minknow 4.0.X"
        )
        log.warning("If you experience problems, let us know.")


class RpcSafeConnection:
    """
    Get values from the RPC and except RPC errors correctly
    """

    def __init__(self, api_connection):
        """

        Parameters
        ----------
        api_connection: minknow_api.Connection
            Connection to the flowcell position
        """
        self.api_connection = api_connection

    def get_run_info(self):
        """
        Get the protocol service run info or None if error
        Returns
        -------
        minknow_api._support.MessageWrapper or None
            Message wrapper for the run information or None if error
        """
        try:
            return self.api_connection.protocol.get_run_info()
        except RpcError:
            log.debug("Failed to get protocol run info")
            return None

    def get_current_protocol_run(self):
        """
        Get thr protocol run from the RPC or return None upon error
        Returns
        -------
        minknow_api._support.MessageWrapper or None
            Message wrapper for the run information or None if error

        """
        try:
            return self.api_connection.protocol.get_current_protocol_run()
        except RpcError as E:
            log.debug("Failed to get current run info from protocol service: {}".format(E))
            return None

    def get_acquisition_data(self):
        """
        Set the acquisition data on the device monitor class
        Parameters
        ----------
        Returns
        -------
        minknow_api._support.MessageWrapper
            The acquisition information for this position
        """
        try:
            return self.api_connection.acquisition.get_acquisition_info()
        except RpcError as e:
            log.debug(e)
            return None

    def get_current_acquisition_run(self):
        """
        get the current acquisition run
        Returns
        -------
        minknow_api._support.MessageWrapper
            The acquisition run information for this position
        """
        try:
            return self.api_connection.acquisition.get_current_acquisition_run()
        except RpcError as e:
            log.debug(e)
            return None

    def get_run_info_sample_name(self):
        """
        Get the run info sample name. or set it to Mux_scan or platform qc
        Returns
        -------

        """
        try:
            sample_id = self.api_connection.protocol.get_run_info().user_info
        except RpcError:
            c = ProtocolRunUserInfo()
            # TODO watch out here as this is fixing any unknown run as Platform QC
            c.sample_id.value = "Mux_scan"
            c.protocol_group_id.value = "Platform_QC"
            sample_id = c
            log.debug("Sample ID not yet known.")
        return sample_id


class LiveMonitoringActions(RpcSafeConnection):
    """
    Actions that we need to take when certain events are triggered in the course of sequencing. Also communicates with
    minoTpur
    """

    def __init__(self, api_connection, minotour_api, sequencing_statistics, header):
        """
        Initialise this
        Parameters
        ----------
        api_connection: minknow_api.Connection
            Connection to the flowcell position
        minotour_api: minFQ.minotourapi.MinotourAPI
            Class for convenience request to minoTOur
        sequencing_statistics: SequencingStatistics
            Track the fastq monitoring metrics and folders with this class
        header: dict
            The dict for request headers
        """
        self.api_connection = api_connection
        self.minotour_api = minotour_api
        self.sequencing_statistics = sequencing_statistics
        self.header = header
        self.long_interval = 30
        super().__init__(self.api_connection)

    def run_stop(self):
        """
        Stop a run - send a post request to minoTour, remove the folder from the sequencing statistics clsas
        Parameters
        ----------
        Returns
        -------

        """
        log.info("Run stop observed for Run - {}".format(str(self.run_information.run_id)))
        self.update_minion_event()
        folder_path = self.folder_path
        log.info("popping folder path {}".format(folder_path))
        if not self.args.no_fastq:
            if folder_path in self.sequencing_statistics.watched_directory_set:
                self.sequencing_statistics.watched_directory_set.remove(folder_path)
            if folder_path in self.sequencing_statistics.to_watch_directory_list:
                self.sequencing_statistics.to_watch_directory_list.remove(folder_path)
            if str(self.run_information.run_id) in self.sequencing_statistics.fastq_info:
                log.info("altering live acquisition status for : {} from fastq info".format(str(self.run_information.run_id)))
                log.info(pformat(self.sequencing_statistics.fastq_info))
                self.sequencing_statistics.fastq_info[str(self.run_information.run_id)]["acquisition_finished"] = True
                log.info("altered")
                log.info(pformat(self.sequencing_statistics.fastq_info))
            self.sequencing_statistics.update = True
        log.debug("run stop"
                  " observed")

    def run_start(self):
        """
        This function will fire when a run first begins.
        Creates the run in the class and update the minion run info in minotour

        Returns
        -------
        None
        """
        self.update_minion_event()
        # We wait for 10 seconds to allow the run to start
        time.sleep(10)
        try:
            self.run_information = self.get_current_acquisition_run()
            # we need this information
            while not self.run_information:
                self.run_information = self.get_current_acquisition_run()
                time.sleep(1)
            self.create_run()
            log.debug("RUn created")
            # Grab the folder and if we are allowed, add it to the watchlist
            FolderPath = (
                self.api_connection.protocol.get_current_protocol_run().output_path
            )
            self.folder_path = os.path.abspath(FolderPath)
            if not self.args.no_fastq:
                if (
                    self.folder_path
                    not in self.sequencing_statistics.to_watch_directory_list
                ):
                    self.sequencing_statistics.to_watch_directory_list.append(
                        self.folder_path
                    )
                    self.sequencing_statistics.fastq_info[
                        str(self.run_information.run_id)
                    ]["directory"] = self.folder_path
                    self.sequencing_statistics.fastq_info[
                        str(self.run_information.run_id)
                    ]["run_id"] = self.run_information.run_id
                    self.sequencing_statistics.fastq_info[
                        str(self.run_information.run_id)
                    ]["acquisition_finished"] = False
            self.update_minion_run_info()
        except Exception as err:
            log.error("Problem starting run {}", err)

    def create_run(self):
        """
        Fired to create a run in Minotour, and return a hyperlinked URL to the database entry, and the run primary key.
        Parameters
        ----------
        Returns
        -------
        None

        """
        run_id = self.run_information.run_id
        run = self.minotour_api.get_json(
            EndPoint.RUNS, base_id=run_id, params="search_criteria=runid"
        )
        if run:
            self.minotour_run_url = run["url"]
            self.run_primary_key = run["id"]
        else:
            # get or create a flowcell
            self.sample_id = self.get_run_info_sample_name()
            flowcell_name = (
                "{}_{}".format(self.get_flowcell_id(), self.sample_id.sample_id.value)
                if self.args.force_unique
                else self.get_flowcell_id()
            )
            flowcell = self.minotour_api.get_json(
                EndPoint.FLOWCELL, base_id=flowcell_name, params="search_criteria=name"
            )["data"]
            if not flowcell:
                log.debug("Manually creating flowcell")
                flowcell = self.minotour_api.post(
                    EndPoint.FLOWCELL, no_id=True, json={"name": flowcell_name}
                )
            payload = {
                "name": self.sample_id.sample_id.value,
                "sample_name": self.sample_id.sample_id.value,
                "runid": run_id,
                "is_barcoded": False,
                "has_fastq": True,
                "flowcell": flowcell["url"],
                "minion": self.minion["url"],
                "start_time": str(
                    self.run_information.start_time.ToDatetime().replace(
                        tzinfo=pytz.UTC
                    )
                ),
            }

            created_run = self.minotour_api.post(
                EndPoint.RUNS, json=payload, no_id=True
            )
            if not created_run:
                log.error("Run not created!")
            else:
                self.minotour_run_url = created_run["url"]
                self.run_primary_key = created_run["id"]  # I

        log.debug("***** self.run_primary_key: {}".format(self.run_primary_key))
        log.debug("**** run stats updated")

    def first_connect(self):
        """
        This function will run when we first connect to the MinION device.
        It will provide the information to minotour necessary to remotely control the minION device.
        Returns
        -------
        bool
            Device activity state.
        """
        log.debug("First connection observed")
        log.debug("Current acquistion status: {}".format(self.acquisition_status))
        self.update_minion_event()
        # set the device to active as it is currently sequencing
        if str(self.acquisition_status) in {
            "ACQUISITION_RUNNING",
            "ACQUISITION_STARTING",
        }:
            self.device_active = True

    def disconnect_nicely(self):
        """
        User has ^C to quit minFQ. This function fires to let minoTour know.
        Returns
        -------

        """
        self.run_bool = False
        self.acquisition_status = "unplugged"
        self.update_minion_event()
        # todo we need to look at this to handle manual unplugged
        self.minion_status["minKNOW_status"] = "unplugged"
        self.minion_status = self.update_minion_info()

    def minknow_command(self):
        """
        This function will recieve commands for a specific minION and handle the interaction.
        :return:
        """
        pass

    def send_message(self, severity_level, message):
        """
        Send a message to be displayed in the minKnow log.
        Parameters
        ----------
        severity_level: int
            The severity level of the message. 0 - low, 1 - medium, 2- high
        message: str
            The message to be sent to the display

        Returns
        -------

        """
        self.api_connection.log.send_user_message(
            severity=severity_level, user_message=message
        )

    def get_flowcell_id(self):
        """
        Get the user specified flowcell ID if there is one, else get the flowcell ID that minKNOW reads from the actual
         flowcell
        Returns
        -------
        str
            Flowcell id
        """
        if self.flowcell_data.user_specified_flow_cell_id:
            flowcell_id = self.flowcell_data.user_specified_flow_cell_id
        else:
            flowcell_id = self.flowcell_data.flow_cell_id
        return flowcell_id

    def get_seq_device_settings(self):
        """
        Get the minion settings for the device based on the device type
        Parameters
        ----------

        Returns
        -------
        minknow_api._support.MessageWrapper
            The device settings for this minion
        """
        if str(self.device_type) == ("PROMETHION"):
            device_settings = (
                self.api_connection.promethion_device.get_device_settings()
            )
        else:
            device_settings = self.api_connection.minion_device.get_settings()
        return device_settings

    def update_minion_info(self):
        """
        Update the minion info concept in MinoTour server. This stores information for this position
         about the connected version of minknow and the computer it is running on.
        Parameters
        ----------

        Returns
        -------

        """
        current_script = (
            self.api_connection.protocol.get_run_info().protocol_id
            if self.acquisition_data
            else "No Run"
        )
        try:
            self.file_system_disk_space = self.api_connection.instance.get_disk_space_info().filesystem_disk_space_info[
                0
            ]
        except RpcError as e:
            log.debug(e)
        file_system_disk_space = self.file_system_disk_space
        payload = {
            "minion": self.minion["url"],
            "minKNOW_status": self.acquisition_status,
            "minKNOW_current_script": current_script,
            "minKNOW_exp_script_purpose": self.api_connection.protocol.get_protocol_purpose().purpose,
            "minKNOW_flow_cell_id": self.get_flowcell_id(),
            "minKNOW_real_sample_rate": self.api_connection.device.get_sample_rate().sample_rate,
            "minKNOW_asic_id": self.flowcell_data.asic_id_str,
            "minKNOW_total_drive_space": file_system_disk_space.bytes_capacity,
            "minKNOW_disk_space_till_shutdown": file_system_disk_space.bytes_when_alert_issued,
            "minKNOW_disk_available": file_system_disk_space.bytes_available,
            "minKNOW_warnings": file_system_disk_space.recommend_stop,
            "minknow_version": self.minknow_version,
            "wells_per_channel": self.flowcell_data.wells_per_channel,
        }

        current_protocol_run = self.get_current_protocol_run()
        if current_protocol_run:
            if current_protocol_run.acquisition_run_ids:
                payload[
                    "minKNOW_script_run_id"
                ] = current_protocol_run.acquisition_run_ids[0]

        if self.sample_id:
            payload["minKNOW_sample_name"] = self.sample_id.sample_id.value
            payload["minKNOW_run_name"] = self.sample_id.protocol_group_id.value

        if self.acquisition_data:
            payload["read_length_type"] = (
                "BASECALLED BASES"
                if self.acquisition_data.config_summary.basecalling_enabled
                else "ESTIMATED BASES"
            )

        if self.run_information:
            if hasattr(self.run_information, "run_id"):
                payload["minKNOW_hash_run_id"] = str(self.run_information.run_id)

        method_handle = (
            self.minotour_api.put if self.minion_status else self.minotour_api.post
        )
        self.minion_status = method_handle(
            EndPoint.MINION_STATUS, base_id=self.minion["id"], json=payload
        )

    def update_minion_event(self):
        """
        Update the event that minoTour has for this minion to match the currecnt acquistion status
        Returns
        -------
        None
        """
        if not self.acquisition_status:
            time.sleep(1)
        for info in self.minotour_api.minion_event_types:
            if info.get("name", "") == self.acquisition_status:
                minknow_status_url = info["url"]
        payload = {
            "computer_name": self.computer_name,
            "datetime": str(datetime.datetime.now()),
            "event": str(urlparse(minknow_status_url).path),
            "minION": str(self.minion["url"]),
        }
        self.minotour_api.post(
            EndPoint.MINION_EVENT, base_id=self.minion["id"], json=payload
        )

    def update_minion_run_info(self):
        """
        Update minion_run_info for this position in Minotour, sent once at the start of the run.
        Returns
        -------

        """
        payload = {
            "minion": self.minion["url"],
            "minKNOW_current_script": self.api_connection.protocol.get_run_info().protocol_id,
            "minKNOW_sample_name": self.sample_id.sample_id.value,
            "minKNOW_exp_script_purpose": self.api_connection.protocol.get_protocol_purpose().purpose,
            "minKNOW_flow_cell_id": self.get_flowcell_id(),
            "minKNOW_run_name": self.sample_id.sample_id.value,
            "run": self.minotour_run_url,
            "minKNOW_version": self.api_connection.instance.get_version_info().minknow.full,
            "minKNOW_hash_run_id": self.run_information.run_id,
            "minKNOW_script_run_id": self.api_connection.protocol.get_current_protocol_run().acquisition_run_ids[
                0
            ],
            "minKNOW_real_sample_rate": self.api_connection.device.get_sample_rate().sample_rate,
            "minKNOW_asic_id": self.flowcell_data.asic_id_str,
            "minKNOW_start_time": self.run_information.start_time.ToDatetime().strftime(
                "%Y-%m-%d %H:%M:%S"
            ),
            "minKNOW_colours_string": MessageToJson(
                self.api_connection.analysis_configuration.get_channel_states_desc(),
                preserving_proto_field_name=True,
                including_default_value_fields=True,
            ),
            "minKNOW_computer": self.computer_name,
            "target_temp": self.temperature_data.target_temperature,
            "flowcell_type": self.flowcell_data.user_specified_product_code,
        }

        payload.update(self.api_connection.protocol.get_context_info().context_info)
        run_info = self.get_run_info()
        if run_info.user_info.HasField("protocol_group_id"):
            payload["experiment_id"] = run_info.user_info.protocol_group_id.value
        else:
            payload["experiment_id"] = "Not Known"
        update_run_info = self.minotour_api.post(
            EndPoint.MINION_RUN_INFO, json=payload, base_id=self.run_primary_key
        )

    def update_minion_stats(self):
        """
        Update the statistics about a run that we have recorded from minKnow.
        Sent to Minotour and stored in minIon run stats table.
        Contains information about the run, not just minKNOW/minION.
        Returns
        -------
        None
        """
        counted = Counter(self.channel_states_dict.values())
        payload = {
            "minion": str(self.minion["url"]),
            "run": self.minotour_run_url,
            "sample_time": str(datetime.datetime.now()),
            "event_yield": self.acquisition_data.yield_summary.selected_events,
            "asic_temp": self.temperature_data.minion.asic_temperature.value,
            "heat_sink_temp": self.temperature_data.minion.heatsink_temperature.value,
            "voltage_value": int(self.bias_voltage),
            "minKNOW_read_count": self.acquisition_data.yield_summary.read_count,
            "estimated_selected_bases": self.acquisition_data.yield_summary.estimated_selected_bases,
        }
        payload.update(counted)
        if self.histogram_data:
            try:
                payload.update(
                    {
                        "minKNOW_histogram_values": str(
                            self.histogram_data.histogram_data[0].bucket_values
                        ),
                        "minKNOW_histogram_bin_width": self.histogram_data.bucket_ranges[
                            0
                        ].end,
                        # ToDo Determine if this actually exsits!
                        "actual_max_val": self.histogram_data.source_data_end,
                        "n50_data": self.histogram_data.histogram_data[0].n50,
                    }
                )
            except IndexError as e:
                log.error(
                    "Problem access histogram data for {}".format(
                        str(self.minion["url"])
                    )
                )
        if self.acquisition_data:
            payload.update(
                {
                    "basecalled_bases": int(
                        self.acquisition_data.yield_summary.basecalled_pass_bases
                    )
                    + int(self.acquisition_data.yield_summary.basecalled_fail_bases),
                    "basecalled_fail_read_count": self.acquisition_data.yield_summary.basecalled_fail_read_count,
                    "basecalled_pass_read_count": self.acquisition_data.yield_summary.basecalled_pass_read_count,
                }
            )
        result = self.minotour_api.post(
            EndPoint.MINION_RUN_STATS, json=payload, base_id=self.run_primary_key
        )



import json
import logging
import operator
import os
import sys
import time
from collections import defaultdict, OrderedDict
import curses
from pprint import pformat

from minFQ.endpoints import EndPoint
from .version import __version__

log = logging.getLogger(__name__)


class CursesHandler(logging.Handler):
    """ todo spruce up message
    Logging handler to emit logs to the curses window in the correct format
    """

    def __init__(self, screen, log_pad):
        """

        Parameters
        ----------
        screen: _curses.window
            The main terminal screen
        log_pad: _curses.window
            The pad that we are printing out log too.
        """
        logging.Handler.__init__(self)
        self.screen = screen
        self.log_pad = log_pad

    def emit(self, record):
        try:
            msg = self.format(record)
            screen = self.screen
            log_pad = self.log_pad
            fs = "\n%s"
            log_pad.addstr(fs % msg)
            # screen.box()
            # refresh_pad(screen, log_pad)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


class SequencingStatistics:
    """
    A class to store data about the sequencing, between different threads.
    """

    def __init__(self):
        self.screen_num_rows = 0
        self.screen_num_cols = 0
        self.files_seen = 0
        self.files_processed = 0
        self.files_skipped = 0
        self.reads_seen = 0
        self.reads_corrupt = 0
        self.reads_skipped = 0
        self.reads_uploaded = 0
        self.fastq_message = "No Fastq Seen"
        self.update = False
        self.read_up_time = time.time()
        self.read_count = 0
        self.to_watch_directory_list = []
        self.watched_directory_set = set()
        self.errored = False
        self.error_message = ""
        self.time_per_file = time.time()
        self.minfq_info = {}
        self.minfq_initialisation_time = time.time()
        self.fastq_info = defaultdict(lambda: defaultdict(int))
        self._connected_positions = {}
        self.minfq_y = 2
        self.minknow_y = 10
        self._fastq_y = 16
        self.minotour_url = ""
        self.minknow_sample_col_x = 67

    @property
    def fastq_y(self):
        """
        Return te number of connected positions to calulcate where to write the fastq statistics
        Returns
        -------

        """
        return self._fastq_y + len(self._connected_positions)

    @property
    def connected_positions(self):
        if self._connected_positions:
            return OrderedDict(
                sorted(
                    {
                        key: {
                            k: self.convert(time.time() - v)
                            if isinstance(v, float)
                            else v
                            for k, v in value.items()
                        }
                        for key, value in self._connected_positions.items()
                    }.items(),
                    key=operator.itemgetter(0),
                )
            )
        else:
            return {}

    def update_minknow_cols_x(self, minknow=True):
        """
        Update the minknow cols x finding the widest value of each column. If more than 64 chars truncate it.
        Returns
        -------
        None or dict
            Returns dict of char width if minknow is false
        """
        a = self.connected_positions if minknow else self.fastq_info_test
        char_widths = {}
        for position_values in a.values():
            char_widths.update(
                {
                    key: len(str(inner_values))
                    for key, inner_values in position_values.items()
                    if len(str(inner_values)) > char_widths.get(key, 0)
                }
            )
        if minknow:
            if char_widths.get("sample_id", 0) > 8:
                self.minknow_sample_col_x = 67 + char_widths["sample_id"] - 8
            else:
                self.minknow_sample_col_x = 67
        else:
            return char_widths

    @property
    def minfq_uptime(self):
        """
        How long this minFQ instance has been up
        Returns
        -------
        str
        """
        return self.convert(time.time() - self.minfq_initialisation_time)

    @property
    def elapsed(self):
        """
        Returns
        -------
        """
        return self.convert(time.time() - self.read_up_time)

    def convert(self, seconds):
        """
        Convert seconds to Hours:Minutes:Seconds format
        Returns
        -------
        str
            Hours:Minutes:Seconds format generated from given second amount
        """
        min, sec = divmod(seconds, 60)
        hour, min = divmod(min, 60)
        return "%d:%02d:%02d" % (hour, min, sec)

    @property
    def per_file(self):
        """
        Time to upload the current file
        Returns
        -------
        str
            Hours:Minutes:Seconds format generated from given second amount since this file upload began
        """
        return self.convert(time.time() - self.time_per_file)

    def check_fastq_info(self):
        """
        Check that a fastq info isn't out of data. Pop an entry if acquistion_finished = True and there an files queued
        Returns
        -------
        None
        """
        iteration_safe_dict = json.loads(json.dumps(self.fastq_info))
        for run_id, info_dict in iteration_safe_dict.items():
            files_queued = (
                info_dict.get("files_seen", 0)
                - info_dict.get("files_processed", 0)
                - info_dict.get("files_skipped", 0)
            )
            if info_dict.get("acquisition_finished", False) and not files_queued:
                log.info(
                    "{} being popped as files queued is {} and acquisition is finished ({})".format(
                        run_id, files_queued, info_dict["acquisition_finished"]
                    )
                )
                self.fastq_info.pop(run_id)
                log.info(pformat(self.fastq_info))


def refresh_pad(screen, pad):
    """

    Parameters
    ----------
    screen: _curses.window
        Main curses we add the pad too
    pad: _curses.window
        Curses window
    sequencing_stats: minFQ.utils.SequencingStatistics
        Sequencing stats class stores info

    Returns
    -------
    None
    """
    num_rows, num_cols = screen.getmaxyx()
    # pad.addstr(0, 0, "{} rows, {} cols".format(num_rows, num_cols))
    pad.refresh(0, 0, 0, 0, num_rows - 1, num_cols - 1)
    # sequencing_stats.screen_num_rows = num_rows
    # sequencing_stats.screen_num_cols = num_cols


def ascii_minotour(stdscr):
    """
    Add an ascii minotour in the top right because why not
    Parameters
    ----------
    stdscr

    Returns
    -------

    """
    stdscr.addstr(0, 92, "     .                .", curses.color_pair(2))
    stdscr.addstr(1, 92, "     -+:...       ..:+=", curses.color_pair(2))
    stdscr.addstr(2, 92, "      :+*##%%%%%%##*+:", curses.color_pair(2))
    stdscr.addstr(3, 92, "    :-=+**.*%%%%#.**++=-.", curses.color_pair(2))
    stdscr.addstr(4, 92, " :*@@@@@@@-*%%%%%.@@@@@@@%=", curses.color_pair(2))
    stdscr.addstr(5, 92, "+@@@@@@@@@%.+%%#-+@@@@@@@@@*", curses.color_pair(2))
    stdscr.addstr(6, 92, "@@@@@%.-*@@@*==+%@@%- =@@@@@", curses.color_pair(2))
    stdscr.addstr(7, 92, ":@@@@#   :@@@@@@@@+   -@@@@=", curses.color_pair(2))
    stdscr.addstr(8, 92, " .+@@@+##-:@@@@@@+.*#=%@@#- ", curses.color_pair(2))
    stdscr.addstr(9, 92, "    =#@@@@ #@@@@@.=@@@%+.   ", curses.color_pair(2))
    stdscr.addstr(10, 92, "       :-.:*%%%%%- -:   ", curses.color_pair(2))
    stdscr.addstr(11, 92, "         +%%+--=#%#-        ", curses.color_pair(2))
    stdscr.addstr(12, 92, "        :##      *%+      ", curses.color_pair(2))
    stdscr.addstr(13, 92, "       .:=*:     +*-.", curses.color_pair(2))


def write_out_minfq_info(stdscr, sequencing_statistics):
    """

    Parameters
    ----------
    stdscr: _curses.window
        curses window
    sequencing_statistics: minFQ.utils.SequencingStatistics
        Class tracking stats about minFQ
    Returns
    -------

    """
    stdscr.addstr(1, 0, "                     ")
    stdscr.addstr(
        sequencing_statistics.minfq_y,
        0,
        "MinFQ Stats\n------------------------------------------\n"
        "MinFQ up time: {}\n"
        "Connected to minoTour at: {}\n".format(
            sequencing_statistics.minfq_uptime, sequencing_statistics.minotour_url
        ),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 4,
        0,
        "Total FASTQ files Queued: {}".format(
            sequencing_statistics.files_seen
            - sequencing_statistics.files_processed
            - sequencing_statistics.files_skipped
        ),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 4,
        43,
        "Total Reads Queued: {}".format(
            sequencing_statistics.reads_seen
            - sequencing_statistics.reads_uploaded
            - sequencing_statistics.reads_skipped
        ),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 5,
        0,
        "Total FASTQ files Uploaded: {}".format(sequencing_statistics.files_processed,),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 5,
        43,
        "Total Reads Uploaded: {}".format(sequencing_statistics.reads_uploaded),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 6,
        0,
        "Total FASTQ files Already seen: {}".format(
            sequencing_statistics.files_skipped
        ),
    )
    stdscr.addstr(
        sequencing_statistics.minfq_y + 6,
        43,
        "Total Reads Uploaded: {} ".format(sequencing_statistics.reads_skipped),
    )


def write_out_minknow_info(stdscr, sequencing_statistics):
    """
    Write out information related to this minFQ instances connection to minKNOW
    Parameters
    ----------
    stdscr: _curses.window
        The window containing the terminal messages
    sequencing_statistics: minFQ.utils.SequencingStatistics
        class tracking the sequencing statistics of the run

    Returns
    -------
    None
    """
    stdscr.addstr(
        sequencing_statistics.minknow_y,
        0,
        "MinKNOW Stats\n---------------------------------------------\nPositions connected: {}\n".format(
            len(sequencing_statistics.connected_positions)
        ),
    )
    cols_y = sequencing_statistics.minknow_y + 4
    sequencing_statistics.update_minknow_cols_x()
    stdscr.clrtobot()
    stdscr.addstr(cols_y, 0, "Position")
    stdscr.addstr(cols_y, 16, "Running")
    stdscr.addstr(cols_y, 32, "Run Status")
    stdscr.addstr(cols_y, 57, "Sample")
    stdscr.addstr(
        cols_y, sequencing_statistics.minknow_sample_col_x, "Connection up Time"
    )
    for index, value in enumerate(sequencing_statistics.connected_positions.values()):
        stdscr.addstr(cols_y + index + 1, 0, value["device_id"])
        stdscr.addstr(cols_y + index + 1, 16, value["running"])
        stdscr.addstr(cols_y + index + 1, 32, value.get("acquisition_status", ""))
        stdscr.clrtoeol()
        stdscr.addstr(cols_y + index + 1, 57, value["sample_id"])
        stdscr.addstr(
            cols_y + index + 1,
            sequencing_statistics.minknow_sample_col_x,
            str(value["up_time"]),
        )

    # "Position \t\t Running \t\t Sample \t\t Connection uptime")


def write_out_fastq_info(stdscr, sequencing_statistics):
    """
    write out base-called data upload metrics
    Parameters
    ----------
    stdscr: _curses.window
        The window containing the terminal messages
    sequencing_statistics: minFQ.utils.SequencingStatistics
        class tracking the sequencing statistics of the run

    Returns
    -------
    None

    """
    stdscr.addstr(
        sequencing_statistics.fastq_y,
        0,
        "Base-called data upload stats\n------------------------------\nDirectories watched: {}".format(
            len(sequencing_statistics.watched_directory_set)
        ),
    )
    stdscr.addstr(
        sequencing_statistics.fastq_y + 3,
        0,
        str(sequencing_statistics.fastq_message)
    )
    cols_y = sequencing_statistics.fastq_y + 5
    stdscr.addstr(cols_y, 0, "Run id")
    stdscr.addstr(cols_y, 15, "Queued")
    stdscr.addstr(cols_y, 30, "Uploaded")
    stdscr.addstr(cols_y, 45, "Already Seen")
    stdscr.addstr(cols_y, 60, "Directory")
    stdscr.addstr(cols_y + 1, 15, "Files/Reads")
    stdscr.addstr(cols_y + 1, 30, "Files/Reads")
    stdscr.addstr(cols_y + 1, 45, "Files/Reads")
    iteration_safe_dict = json.loads(json.dumps(sequencing_statistics.fastq_info))
    for index, value in enumerate(iteration_safe_dict.values()):
        stdscr.addstr(cols_y + index + 2, 0, str(value.get("run_id", ""))[:10])
        files_queued = (
            value.get("files_seen", 0)
            - value.get("files_processed", 0)
            - value.get("files_skipped", 0)
        )
        reads_queued = (
            value.get("reads_seen", 0)
            - value.get("reads_uploaded", 0)
            - value.get("reads_skipped", 0)
        )
        stdscr.addstr(
            cols_y + index + 2, 15, "{}/{}".format(str(files_queued), str(reads_queued))
        )
        stdscr.addstr(
            cols_y + index + 2,
            30,
            "{}/{}".format(
                str(value.get("files_processed", 0)),
                str(value.get("reads_uploaded", 0)),
            ),
        )
        stdscr.addstr(
            cols_y + index + 2,
            45,
            "{}/{}".format(
                str(value.get("files_skipped", 0)), str(value.get("reads_skipped", 0))
            ),
        )
        stdscr.addstr(cols_y + index + 2, 60, str(value.get("directory", 0)))


def test(stdscr):
    curses.nocbreak()
    stdscr.keypad(False)
    curses.echo()
    curses.curs_set(1)
    curses.endwin()
    return __version__


def add_arguments_to_parser(parser, stdscr):
    """
    Add command line arguments to the parser.
    Parameters
    ----------
    parser: configargparse.ArgumentParser
        The argument parser instance
    stdscr: _curses.window
        The curses window object
    Returns
    -------
    parser: configargparse.ArgumentParser
        The parser with the added arguments
    """

    parser.add_argument(
        "-hn",
        "--hostname",
        type=str,
        # required=True,
        default="127.0.0.1",
        help="The host name for the minoTour server.",
        dest="host_name",
    )

    parser.add_argument(
        "-p",
        "--port",
        type=int,
        # required=True,
        default=80,
        help="The port number for the minoTour server.",
        dest="port_number",
    )

    parser.add_argument(
        "-k",
        "--key",
        type=str,
        required=True,
        default=None,
        help="The api key for uploading data.",
        dest="api_key",
    )

    parser.add_argument(
        "-w",
        "--watch-dir",
        type=str,
        # required=True,
        default=None,
        help="The path to the folder containing the downloads directory with fast5 reads to analyse - e.g. C:\\data\\minion\\downloads (for windows).",
        dest="watch_dir",
    )

    parser.add_argument(
        "-i",
        "--ignore_existing",
        action="store_true",
        required=False,
        default=False,
        help="The client will ignore previously existing fastq files and will only monitor newly created files..",
        dest="ignore_existing",
    )

    parser.add_argument(
        "-s",
        "--skip_sequence",
        action="store_true",
        required=False,
        help="If selected only read metrics, not sequence, will be uploaded to the databse.",
        dest="skip_sequence",
    )

    parser.add_argument(
        "-nf",
        "--no_fastq",
        action="store_true",
        help="Run minFQ without monitoring fastq files.",
        default=False,
        dest="no_fastq",
    )

    parser.add_argument(
        "-nm",
        "--no_minKNOW",
        action="store_true",
        help="Run minFQ without monitoring minKNOW for live activity.",
        default=False,
        dest="no_minknow",
    )

    parser.add_argument(
        "-rc",
        "--remote_control",
        action="store_true",
        default=False,
        help="This option allows your runs to be remotely started and stopped and for runs to be remotely renamed. As standard this is not enbabled.",
        dest="enable_remote",
    )

    parser.add_argument(
        "-ip",
        "--ip-address",
        type=str,
        dest="ip",
        required=False,
        default="127.0.0.1",
        help="The IP address of the minKNOW machine - Typically 127.0.0.1.",
    )

    parser.add_argument(
        "-n",
        "--name",
        type=str,
        default=None,
        help="This provides a backup name for a flowcell. MinoTour will use the run names and flowcell ids it finds in reads or from minKNOW if available.",
        dest="run_name_manual",
    )

    parser.add_argument(
        "--unique",
        action="store_true",
        default=True,
        help="If you are flushing a flowcell, this option will force the flowcell to be named as a combination of flowcell ID and sample name. Thus data will be grouped appropriately. Default true.",
        dest="force_unique",
    )

    parser.add_argument(
        "-f",
        "--is_flowcell",
        action="store_true",
        help="If you add this flag, all runs added here will be considered as a single flow cell with the name set by the name flag.",
        dest="is_flowcell",
    )

    parser.add_argument(
        "-tc",
        "--treatment-control",
        type=int,
        required=False,
        default=None,
        help="Optionally split reads based in treatment and control groups based on the channel number. The integer value informed is used to mover ish read to the control group.",
        dest="treatment_control",
    )

    parser.add_argument(
        "-j",
        "--job",
        type=int,
        # required=True,
        default=None,
        help="An optional minotour job to run on your server. Please enter the ID shown on the side when running --list.",
        dest="job",
    )

    parser.add_argument(
        "-r",
        "--reference",
        type=int,
        # required=True,
        default=None,
        help="An optional minotour reference to map against. please enter the numerical id shown when running --list",
        dest="reference",
    )

    parser.add_argument(
        "--list",
        action="store_true",
        required=False,
        help="List available tasks, target sets and references at this server.",
        dest="list",
    )

    parser.add_argument(
        "-ts",
        "--targets",
        type=int,
        default=None,
        help="Set the target set for the metagenomics, if desired. Please enter the numerical id shown when running --list",
        dest="targets",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Don't clear the screen. Helps when debugging.",
        default=False,
        dest="verbose",
    )

    parser.add_argument(
        "-ll",
        "--loglevel",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level, default INFO.",
        dest="loglevel",
    )

    parser.add_argument(
        "-V", "--version", action="version", version=test(stdscr),
    )

    parser.add_argument(
        "-T",
        "--toml",
        default=None,
        required=False,
        help="Path to the channels configuration file for a read until experiment.",
        dest="toml",
    )
    parser.add_argument(
        "-U",
        "--unblocks",
        default=None,
        required=False,
        help="Absolute path to an unblocked read_ids text file. Not necessary, should be picked up automatically.",
        dest="unblocks",
    )
    parser.add_argument(
        "-ps",
        "--primer-scheme",
        type=int,
        default=None,
        required=False,
        help="Set the primer scheme to use for artic tasks. Valid options can be seen using --list.",
        dest="primer_scheme"
    )
    parser.add_argument(
        "-cc",
        "--concat_conditions",
        default=False,
        required=False,
        help="On ReadFish runs, concatenate the reads barcode name and the condition ReadFish has for it's channel. "
             "Default False. ",
        dest="concat_condish",
        action="store_true",
    )
    return parser


def validate_args(args):
    """
    Check the args before setting up all connections. Error out if any mistakes are found.
    Parameters
    ----------
    args: argparse.Namespace
        Namespace for chosen arguments after parsing
    Returns
    -------
    None
    """
    error_string = None
    if args.watch_dir:
        if not os.path.exists(args.watch_dir):
            error_string = "The watch directory specified {} does not exist. Please check specified path.".format(
                args.watch_dir
            )

        if args.no_fastq:
            error_string = (
                "If not monitoring FASTQ please do not provide a watch directory."
            )
    if args.job and args.job == 16:
        if not args.reference:
            error_string = "Reference required for Artic job type. Please provide."
        if not args.primer_scheme:
            error_string = "Primer Scheme required for Artic job type. Please provide."

    if args.toml is not None:
        if not os.path.exists(args.toml):
            error_string = "Toml file not found in this location. Please check that the specified file path is correct."

    if args.unblocks is not None:
        if not os.path.exists(args.unblocks):
            error_string = (
                "Unblocked read ids file not found in this location."
                " Please check that the specified file path is correct."
            )
    if args.no_fastq and args.no_minknow:
        error_string = (
            "You must monitor either FastQ or MinKNOW.\n This program will now exit."
        )
    if not args.no_minknow and args.ip is None:
        error_string = (
            "To monitor MinKNOW in real time you must specify the IP address"
            " of your local machine.\nUsually: -ip 127.0.0.1"
        )

    if error_string:
        raise Exception(error_string)


def check_server_compatibility(minotour_api, log):
    """
    Check the minoTOur servers compatibility with minFQ
    Parameters
    ----------
    minotour_api: minFQ.minotourapi.MinotourAPI
        Minotour api connection Class
    log: logging.Logger
        Logger for this script
    Returns
    -------
    str
    """
    version = minotour_api.get_json(EndPoint.VERSION)
    clients = version["clients"]
    if not any([__version__.startswith(client) for client in clients]):
        log.error(
            "Server does not support this client. Please check server version or upgrade/downgrade client."
        )
        log.error("Server supports minFQ versions:")
        for client in clients:
            log.error(client)
        raise Exception(
            "Server does not support this client. "
            f"Please change your client to any of {clients} or upgrade your minoTour server."
        )
    else:
        return "minFQ version: {} is compatible with minoTour server specified.\n".format(
            __version__
        )


def list_minotour_options(log, args, minotour_api, stdscr, screen):
    """
    List the options for jobs we can start, references and threat sets for metagenomics
    Parameters
    ----------
    log: logging.Logger
        The logger for this script
    args: argparse.Namespace
        Namespace for chosen arguments after parsing
    minotour_api: minFQ.minotourapi.MinotourAPI
        API convenience class
    stdscr: _curses.pad
        The window for the curses display
    screen: _curses.window
        The window for the curses display
    Returns
    -------
    None
    """
    stdscr.clear()
    stdscr.addstr(
        0, 0, "Fetching available Jobs, References and Targets...", curses.color_pair(5)
    )
    # TODO combine below into new single API end point
    jobs = minotour_api.get_json(EndPoint.TASK_TYPES, params={"cli": True})["data"]
    references = minotour_api.get_json(EndPoint.REFERENCES)["data"]
    schemes = minotour_api.get_json(EndPoint.PRIMER_SCHEMES)["data"]
    params = {"api_key": args.api_key, "cli": True}
    targets = minotour_api.get_json(EndPoint.TARGET_SETS, params=params)
    stdscr.addstr(
        2, 0, "The following jobs are available on this minoTour installation:"
    )
    line_num = 3
    for job_info in jobs:
        if not job_info["name"].startswith("Delete"):
            stdscr.addstr(
                line_num,
                0,
                "\t{}:{}".format(
                    job_info["id"], job_info["name"].lower().replace(" ", "_")
                ),
                curses.color_pair(3),
            )
            line_num += 1
    line_num += 1
    stdscr.addstr(
        line_num,
        0,
        "If you wish to run an alignment, the following references are available:",
    )
    line_num += 1
    for reference in references:
        stdscr.addstr(
            line_num,
            0,
            "\t{}:{}".format(reference["id"], reference["name"]),
            curses.color_pair(4),
        )
        line_num += 1
    line_num += 1
    stdscr.addstr(
        line_num,
        0,
        "If you wish to add a target set to the metagenomics task, the following sets are available to you:",
    )
    line_num += 1
    index = 1
    for target in targets:
        stdscr.addstr(
            line_num, 0, "\t{}:{}".format(index, target), curses.color_pair(2)
        )
        index += 1
        line_num += 1
    line_num += 1
    stdscr.addstr(
        line_num,
        0,
        "Artic primer scheme choice - MANDATORY for starting artic Jobs - use the ID number on the right",
    )
    line_num += 1
    for scheme in schemes:
        stdscr.addstr(
            line_num, 0, "\t{} : {} - {}".format(scheme["id"], scheme["scheme_species"], scheme["scheme_version"]), curses.color_pair(2)
        )
        index += 1
        line_num += 1
    refresh_pad(screen, stdscr)
    curses.napms(6000)
    curses.nocbreak()
    stdscr.keypad(False)
    curses.echo()
    curses.curs_set(1)
    curses.endwin()
    sys.exit(0)


def check_job_from_client(args, log, minotour_api, parser):
    """
    Check that we can start the specified job on the server.
    Parameters
    ----------
    args: argparse.Namespace
        Namespace for chosen arguments after parsing
    log: logging.Logger
        The logger for this script
    minotour_api: minFQ.minotourapi.MinotourAPI
        API class for connecting to minoTour
    parser: configargparse.ArgumentParser
        command line argument parser
    Returns
    -------
    """
    args.job = int(args.job)
    # Get availaible jobs
    jobs = minotour_api.get_json(EndPoint.TASK_TYPES, params={"cli": True})["data"]
    jobs = [job["id"] for job in jobs]
    if args.job not in jobs:
        parser.error(
            "Can't find the job type chosen. Please double check that it is the same ID shown by --list."
        )

    if args.job == "minimap2" or args.job in {4, 5}:
        if args.reference == None:
            log.error("You need to specify a reference for a Minimap2 task.")
            sys.exit(0)
        references = minotour_api.get_json(EndPoint.REFERENCES, params={"cli": True})
        if args.reference not in [reference["id"] for reference in references["data"]]:
            log.error("Reference not found. Please recheck.")
            sys.exit(0)
    if args.job == "metagenomics" or args.job == 10:
        if args.targets:
            if not isinstance(args.targets, int):
                sys.exit("Please use the numerical identifier for the target set.")
            params = {"api_key": args.api_key, "cli": True}
            targets = minotour_api.get_json(EndPoint.TARGET_SETS, params=params)
            if args.targets > len(targets):
                log.error("Target set not found. Please check spelling and try again.")
                sys.exit(0)
    if args.job == "artic" or args.job == 16:
        if not args.primer_scheme:
            log.error("Primer scheme required for artic job.")
            sys.exit(1)


import logging
import threading
import time


from grpc import RpcError
import minknow_api
from minknow_api.manager import Manager
from minknow_api.acquisition_pb2 import AcquisitionState
from minknow_api.device import get_device_type

from minFQ.endpoints import EndPoint
from minFQ.minotourapi import MinotourAPI
from minFQ.minknow_connection_utils import (
    check_warnings,
    LiveMonitoringActions,
)
from minFQ.utils import CursesHandler

log = logging.getLogger(__name__)


class DeviceMonitor(LiveMonitoringActions):
    """
    Monitor a flowcell position, for sequencing metrics
    """

    def __init__(
        self, args, api_connection, header, position_id, sequencing_statistics
    ):
        """

        Parameters
        ----------
        args: argparse.NameSpace
            The argument name space
        api_connection: minknow_api.Connection
            Connection to the flowcell position
        header: dict
            The header to set on the request to the server
        position_id: str
            The name of the position, like MS0000 or X1
        sequencing_statistics: minFQ.utils.SequencingStatistics
            The metrics for the fastq uploading
        """
        self.args = args
        self.minotour_api = MinotourAPI(
            self.args.host_name, self.args.port_number, header
        )
        # Set a status to hold what we are currently doing for this device.
        self.device_active = False
        self.api_connection = api_connection
        self.minknow_version = (
            self.api_connection.instance.get_version_info().minknow.full
        )
        self.device_type = get_device_type(self.api_connection).name
        # Here we need to check if we are good to run against this version.
        check_warnings(self.device_type, self.minknow_version)
        self.header = header
        self.channel_count = (
            self.api_connection.device.get_flow_cell_info().channel_count
        )  # this is ok
        self.channel_states_dict = {i: None for i in range(1, self.channel_count + 1)}
        # self.status = AcquisitionState.ACQUISITION_COMPLETED
        self.acquisition_status = ""
        self.interval = 30  # we will poll for updates every 30 seconds.
        self.long_interval = 30  # we have a short loop and a long loop
        self.position_id = position_id  # This has been renamed from self.minIONid
        self.computer_name = (
            self.api_connection.instance.get_machine_id().machine_id
        )  # This isn't what we want - it reports Macbook-Pro for my computer but should report "DestroyerofWorlds" or similar.

        self.minotour_api.test()
        # TODO 4.0 brings in streaming so we can stream a lot of these in threads instead
        # Updated in Flowcell monitor thread
        self.flowcell_data = self.api_connection.device.get_flow_cell_info()
        # Get minotour minion record or create it if we don't have one
        # todo bug here i'm quite sure for promethion
        self.minion = self.minotour_api._get(EndPoint.GET_MINION,
                                                 params="search_criteria=name",
                                                 base_id=self.position_id)
        if self.minion:
            self.minion = self.minion.json()
        else:
            self.minion = self.minotour_api.post(EndPoint.POST_MINION,
                                   json={"minION_name": self.position_id, "name": self.position_id},
                                   no_id=True
                                   )
        log.info(self.minion)
        self.minion_status = self.minotour_api.get_json(
            EndPoint.MINION_STATUS, base_id=self.minion["id"]
        )
        self.minotour_run_url = ""
        self.run_primary_key = ""
        self.run_bool = True
        self.acquisition_data = self.get_acquisition_data()
        # initialise histogram data
        self.histogram_data = None
        # initialise run information
        self.run_information = None
        self.base_calling_enabled = None
        # Set in the run_info
        self.temperature_data = None
        self.sample_id = None
        self.minion_settings = None
        self.bias_voltage = None
        self.run_info_api = None
        self.folder_path = None
        super().__init__(
            api_connection, self.minotour_api, sequencing_statistics, header
        )

        thread_targets = [
            self.run_monitor_thread,
            self.flowcell_monitor_thread,
            self.run_information_monitor_thread,
            self.get_messages_thread,
            self.channel_state_monitor_thread,
            self.histogram_monitor_thread,
            self.jobs_monitor_thread,
        ]
        for thread_target in thread_targets:
            threading.Thread(target=thread_target, args=(), daemon=True).start()
        self.first_connect()

    def jobs_monitor_thread(self):
        """
        Check the minoTour server for jobs to be sent to minKnow and perform them.
        Returns
        -------

        """
        while self.run_bool:
            data = self.minotour_api.get_json(
                EndPoint.MINION_CONTROL, base_id=self.minion["id"]
            )
            time.sleep(self.interval)
            for job in data:
                if job["job"] == "test_message":
                    self.send_message(
                        1,
                        "minoTour is checking communication status with "
                        + str(self.minion["name"])
                        + ".",
                    )
                elif job["job"] == "custom_message":
                    self.send_message(1, "minoTour: {}".format(job["custom"]))
                elif job["job"] == "stop_minion":
                    if self.args.enable_remote:
                        self.api_connection.protocol.stop_protocol()
                        self.send_message(
                            3, "minoTour was used to remotely stop your run."
                        )
                elif job["job"] == "rename":
                    if self.args.enable_remote:
                        self.api_connection.protocol.set_sample_id(
                            sample_id=job["custom"]
                        )
                        self.send_message(
                            1, "minoTour renamed your run to {}".format(job["custom"])
                        )
                elif job["job"] == "name_flowcell":
                    if self.args.enable_remote:
                        self.api_connection.device.set_user_specified_flow_cell_id(
                            id=job["custom"]
                        )
                        self.send_message(
                            1,
                            "minoTour renamed your flowcell to {}".format(
                                job["custom"]
                            ),
                        )
                elif job["job"] == "start_minion":
                    if self.args.enable_remote:
                        self.api_connection.protocol.start_protocol(
                            identifier=job["custom"]
                        )
                        self.send_message(
                            2, "minoTour attempted to start a run on your device."
                        )
                else:
                    continue
                self.minotour_api.post(
                    EndPoint.MINION_CONTROL,
                    base_id=self.minion["id"],
                    append_id=job["id"],
                    json={},
                )

    def flowcell_monitor_thread(self):
        """
        Monitor flowcell information for this position, and update the class value stored for it.
        Returns
        -------

        """
        while self.run_bool:
            flowcell_info_response = self.api_connection.device.stream_flow_cell_info()
            for flowcell_info in flowcell_info_response:
                log.info("Hello")
                self.flowcell_data = flowcell_info
                self.update_minion_info()

    def histogram_monitor_thread(self):
        """
        Monitor the histogram stream output from minKnow. Set the data on the class. If a run is ongoing, we stay in the
        stream for loop. If a run is not, we check for one every 30 seconds until we go back in the stream loop
        Returns
        -------
        None

        """
        while self.run_bool:
            if str(self.acquisition_status) in {
                "ACQUISITION_STARTING",
                "ACQUISITION_RUNNING",
            }:
                # get the purpose of this sequencing
                purpose = self.api_connection.protocol.get_protocol_purpose().purpose
                if purpose == "sequencing_run":
                    # We need to test if we are doing basecalling or not.
                    self.run_information = self.get_current_acquisition_run()
                    self.base_calling_enabled = False
                    if self.run_information:
                        self.base_calling_enabled = (
                            self.run_information.config_summary.basecalling_enabled
                        )
                    rltype = 2 if self.base_calling_enabled else 1
                    try:
                        histogram_stream = self.api_connection.statistics.stream_read_length_histogram(
                            read_length_type=rltype,
                            bucket_value_type=1,
                            acquisition_run_id=self.run_information.run_id,
                        )
                        for histogram_event in histogram_stream:
                            self.histogram_data = histogram_event
                            if self.acquisition_status not in {
                                "ACQUISITION_RUNNING",
                                "ACQUISITION_STARTING",
                            }:
                                break
                    except Exception as e:
                        log.error("histogram problem: {}".format(e))
                        continue
            time.sleep(self.interval)

    def channel_state_monitor_thread(self):
        """
        Thread to monitor the channel states as put out by minknow. Sets them on the class where they are uploaded
        by a call to update_minion_run_stats somewhere else.
        Returns
        -------
        None

        """
        while self.run_bool:
            if str(self.acquisition_status) in {
                "ACQUISITION_RUNNING",
                "ACQUISITION_STARTING",
            }:
                purpose = self.api_connection.protocol.get_protocol_purpose().purpose
                if purpose == "sequencing_run":
                    last_channel = self.flowcell_data.channel_count
                    channel_states = self.api_connection.data.get_channel_states(
                        wait_for_processing=True,
                        first_channel=1,
                        last_channel=last_channel,
                    )
                    try:
                        for state in channel_states:
                            for channel in state.channel_states:
                                self.channel_states_dict[
                                    int(channel.channel)
                                ] = channel.state_name
                    except Exception as e:
                        log.error("Chan state error: {}".format(e))
            time.sleep(self.interval)

    def run_monitor_thread(self):
        """
        Monitor whether or not a run has just started or stopped. Alerts Minotour in the event of run start/stop.
        Returns
        -------
        None
        """
        while self.run_bool:
            for (
                acquisition_status
            ) in self.api_connection.acquisition.watch_current_acquisition_run():
                self.acquisition_status = AcquisitionState.Name(
                    acquisition_status.state
                )
                self.sequencing_statistics._connected_positions[self.position_id]["acquisition_status"] = str(self.acquisition_status)
                if not self.device_active and str(self.acquisition_status) in {
                    "ACQUISITION_RUNNING",
                    "ACQUISITION_STARTING",
                    "ACQUISITION_FINISHING",
                }:
                    self.device_active = True
                    self.run_start()
                    self.sequencing_statistics._connected_positions[self.position_id]["sample_id"] = str(self.sample_id.sample_id.value)
                    self.sequencing_statistics._connected_positions[self.position_id]["up_time"] = time.time()
                if (
                    self.device_active
                    and str(self.acquisition_status) == "ACQUISITION_COMPLETED"
                ):
                    self.device_active = False
                    self.run_stop()
                    self.sequencing_statistics._connected_positions[self.position_id]["sample_id"] = ""
                log.debug(self.acquisition_status)

    def run_information_monitor_thread(self):
        """
        Get information on the run via the minKnow RPC. Basically just sets class values for any data we can get.
        Returns
        -------
        None

        """
        while self.run_bool:
            self.acquisition_data = self.get_acquisition_data()
            self.temperature_data = self.api_connection.device.get_temperature()
            # ToDo: Check this code on a promethION.
            self.bias_voltage = (
                self.api_connection.device.get_bias_voltage().bias_voltage
            )
            self.run_info_api = self.get_run_info()
            self.sample_id = self.get_run_info_sample_name()
            self.update_minion_info()
            if str(self.acquisition_status) in {
                "ACQUISITION_RUNNING",
                "ACQUISITION_STARTING",
            }:
                self.run_information = (
                    self.api_connection.acquisition.get_current_acquisition_run()
                )
                if self.run_primary_key:
                    self.update_minion_stats()
            time.sleep(self.interval)

    def get_messages_thread(self):
        """
        Get messages from the minKnow client for this flow cell. Include message seen from this flowcell but in older runs.
        Returns
        -------

        """
        while self.run_bool:
            if not self.minotour_run_url:
                time.sleep(5)
                continue
            messages = self.api_connection.log.get_user_messages(
                include_old_messages=True
            )
            for message in messages:
                payload = {
                    "minion": self.minion["url"],
                    "message": message.user_message,
                    "identifier": message.identifier,
                    "severity": message.severity,
                    "timestamp": message.time.ToDatetime().strftime(
                        "%Y-%m-%d %H:%M:%S.%f"
                    )[:-3],
                    "run": self.minotour_run_url,
                }
                self.minotour_api.post(
                    EndPoint.MINION_MESSAGES, base_id=self.minion["id"], json=payload
                )


class MinionManager(Manager):
    """
    manager for flowcell positions listed by minKNOW
    """

    def __init__(
        self,
        **kwargs
    ):
        """
        host="localhost",
        port=None,
        use_tls=False,
        args=None,
        header=None,
        sequencing_statistics=None,
        Parameters
        ----------
        host: str
            The host that minknow server is being served on
        port: int or None
            Default None. IF None, will be defaulted to 9501, the GRPC port for minKNow
        use_tls: bool
            Default False. Use TLS connection
        args: argparse.NameSpace
            The command line arguments after parsing
        header: dict
            The header for the requests to minoTour server - for authorisation
        sequencing_statistics: minFQ.utils.SequencingStatistics
            The metrics for tracking fastq sequencing
        """
        log.info(f"Installed minknow API Version {minknow_api.__version__}")
        if minknow_api.__version__.startswith("5"):
            host = kwargs.get("host", "localhost")
            port = kwargs.get("port", 9502)
            super().__init__(host, port)
        elif minknow_api.__version__[:2] in {"4.2", "4.3", "4.4", "4.5"}:
            host = kwargs.get("host", "localhost")
            port = kwargs.get("port", 9501)
            super().__init__(host, port, use_tls=False)
        self.monitor = True
        self.connected_positions = {}
        self.device_monitor_thread = threading.Thread(
            target=self.monitor_devices, args=()
        )
        self.device_monitor_thread.daemon = True
        self.device_monitor_thread.start()
        self.args = kwargs["args"]
        self.header = kwargs["header"]
        self.sequencing_statistics = kwargs["sequencing_statistics"]

    @property
    def count(self):
        """Count of connected sequencing flowcells listed in the manager"""
        return len(self.connected_positions)

    def monitor_devices(self):
        """
        Monitor devices to see if they are active. If active initialise device monitor class for each flow cell position.
        Returns
        -------

        """
        while self.monitor:
            for position in self.flow_cell_positions():
                # print (position)
                device_id = position.name
                if device_id not in self.connected_positions:
                    running = "yes" if position.running else "no"
                    self.sequencing_statistics._connected_positions[device_id] = {
                        "device_id": device_id,
                        "running": running,
                        "sample_id": "",
                        "up_time": time.time(),
                    }
                    if position.running:
                        connected_position = position.connect()
                        device_connected = connected_position.device.get_device_state().device_state
                        if device_connected:
                            # TODO note that we cannot connect to a remote instance without an ip websocket
                            self.connected_positions[device_id] = DeviceMonitor(
                                self.args,
                                connected_position,
                                self.header,
                                device_id,
                                self.sequencing_statistics,
                            )
                        else:
                            self.sequencing_statistics._connected_positions[device_id]['running']="Error"


            time.sleep(5)

    def stop_monitoring(self):
        """
        Inform minoTour that we have interrupted monitoring, and disconnect from minoTour
        Stops device monitor for each position.
        Returns
        -------
        None
        """
        for flowcell_name, connection in self.connected_positions.items():
            log.info("Disconnecting {} from the server.".format(flowcell_name))
            connection.disconnect_nicely()
        self.monitor = False



import gzip
import json as json_library
import logging
from pprint import pformat

import requests

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from minFQ.Errors import MTConnectionError
from minFQ.endpoints import EndPoint


retry_strategy = Retry(
    total=6,
    status_forcelist=[429, 502, 503, 504],
    method_whitelist=["HEAD", "GET", "OPTIONS", "POST"],
    backoff_factor=1
)
adapter = HTTPAdapter(max_retries=retry_strategy)
http = requests.Session()
http.mount("https://", adapter)
http.mount("http://", adapter)

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

class MinotourAPI:
    def __init__(self, base_url, port_number, request_headers):
        self.base_url = base_url
        self.port_number = port_number
        self.request_headers = request_headers
        self.check_url()
        self.minion_event_types = self.get_json(EndPoint.MINION_EVENT_TYPES)

    def check_url(self):
        """
        Check URL to see if we have a https connection or a http connection.
        Returns
        -------
        None
        """
        if self.base_url.startswith("http://"):
            self.base_url = self.base_url[7:]
        if self.base_url.startswith(("https://")):
            self.base_url = self.base_url[8:]
        if int(self.port_number) != 80:
            r = requests.get("http://{}:{}/".format(self.base_url, self.port_number))
        else:
            r = requests.get("http://{}/".format(self.base_url))
        # print (r.url)
        if r.url.startswith("https"):
            self.base_url = "https://{}/".format(self.base_url)
        else:
            self.base_url = "http://{}:{}/".format(self.base_url, self.port_number)

    def test(self):
        """
        Connect to the server and check everything is cool
        Returns
        -------

        """
        data = self.get(EndPoint.TEST)
        return "Tested connection -> {}\n".format(data)

    def _get(self, endpoint, params=None, **kwargs):
        """
        Get the response to a request to the minoTour server
        Parameters
        ----------
        endpoint:  <enum 'EndPoint'>
            The Enum for the endpoint we wish to get from
        params: dict
            The get request params to include
        base_id: str
            The base id for the url ex. /minion/*1*/
        append_id: str
            The appended id after the base id ex. /minion/1/jobs/*2*
        no_id: bool
            Strip the id space from the url

        Returns
        -------
        requests.models.Response
            The requests object from the request

        """

        url = "{}api/v1{}".format(self.base_url, endpoint.resolve_url(**kwargs))
        resp = http.get(url, headers=self.request_headers, params=params)
        # handle resp fail ...
        self.handle_response(resp)
        return resp

    def get(self, *args, **kwargs):
        """
        Perform get AJAX requests to minoTour server
        Parameters
        ----------
        args
            Expanded function arguments
        kwargs
            Expanded keyword arguments
        Returns
        -------
        str
            The string data response to the request

        """
        return self._get(*args, **kwargs).text

    def get_json(self, *args, **kwargs):
        """
        Get Json from minoTour
        Parameters
        ----------
        args
        kwargs

        Returns
        -------
        dict or list
            Json parsed data string

        """
        # TODO careful as this may not tells us we have errors
        try:
            return json_library.loads(self.get(*args, **kwargs))
        except json_library.JSONDecodeError:
            return ""

    def _post(self, endpoint, json, params=None, **kwargs):
        """
        Perform post AJAX requests to minoTour server
        Parameters
        ----------
        endpoint: str or  <enum 'EndPoint'>
            The partial Url to append the the server address.
        json: dict
            Json str to send containing any data to post.
        params: dict
            Post request parameters
        base_id: str
            The base id for the url ex. /minion/*1*/
        append_id: str
            The appended id after the base id ex. /minion/1/jobs/*2*
        no_id: bool
            If the url has no slug id in it
        Returns
        -------
        requests.models.Response

        """
        url = "{}api/v1{}".format(self.base_url, endpoint.resolve_url(**kwargs))
        json = gzip.compress(json_library.dumps(json).encode())
        resp = http.post(
            url, headers=self.request_headers, data=json, params=params
        )
        self.handle_response(resp)
        return resp

    def post(self, *args, **kwargs):
        """

        Parameters
        ----------
        args
            Expanded arguments
        kwargs
            Expanded Keyword args
        Returns
        -------
        dict or list or str
            Parsed JSON str response, or string response if returned text is not JSON
        """
        resp = self._post(*args, **kwargs)
        log.info("response is {}".format(resp.text))
        try:
            return resp.json()
        except json_library.JSONDecodeError:
            return resp.text

    def _put(self, endpoint, json, params=None, **kwargs):
        """
        perform a put AJAX request to the server.
        Parameters
        ----------
        endpoint: minFQ.endpoints.EndPoint
            Enum containing the url ending
        json: dict
            Dictionary of data to put
        params: str or dict
            Request query string parameters and body parameters
        base_id: str
            The base id for the url ex. /minion/*1*/
        append_id: str
            The appended id after the base id ex. /minion/1/jobs/*2*
        no_id: bool
            Strip the slug format from the enum value string

        Returns
        -------
        requests.models.Response
        """
        url = "{}api/v1{}".format(self.base_url, endpoint.resolve_url(**kwargs))
        json = gzip.compress(json_library.dumps(json).encode())
        resp = http.put(url, headers=self.request_headers, data=json, params=params)
        self.handle_response(resp)
        return resp

    def put(self, *args, **kwargs):
        """
        Perform put AJAX requests to minoTour server
        Parameters
        ----------
        args
        kwargs
        Returns
        -------
        str or dict
            Dict if response is JSON parseable else empty string
        """
        try:
            return self._put(*args, **kwargs).json()
        except json_library.JSONDecodeError:
            return ""

    def delete(self, partial_url, json, parameters=None):
        """
        Perform delete AJAX requests to minoTour server
        Parameters
        ----------
        partial_url: str
            The partial Url to append the the server address.
        json: dict
            The JSON string to send to the server
        parameters: dict
            Put request parameters
        Returns
        -------
        requests.models.Response

        """
        url = "{}api/v1{}?{}".format(self.base_url, partial_url, parameters)
        return requests.delete(
            url, headers=self.request_headers, json=json, params=parameters
        )

    def handle_response(self, response):
        """
        Handle responses as provided by the minFQ client
        Returns
        -------
        None
        """
        if response.status_code not in {200, 201, 204, 400, 404, 502}:
            log.debug(
                "{} responded with status {}".format(response.url, response.status_code)
            )
            log.debug("Text {}".format(response.text))
            log.error(MTConnectionError(response), exc_info=True)
        if response.status_code == 502:
            log.info(
                "Received bad gateway 502 on request {}. Trying to continue....".format(
                    response.url
                )
            )
        return None

    # def get_or_create(self, *args, **kwargs):
    #     """
    #     Get an object from the minoTour server or create it if we receive a 404
    #     Parameters
    #     ----------
    #     args
    #     kwargs
    #
    #     Returns
    #     -------
    #
    #     """
    #     payload = kwargs.pop("json")
    #     resp = self._get(*args, **kwargs)
    #     if not resp:
    #         kwargs["params"] = None
    #         kwargs["json"] = payload
    #         return self.post(*args, **kwargs)
    #     else:
    #         return resp.json()



import gzip
import logging
import os
from pathlib import Path

import numpy as np

from minFQ.endpoints import EndPoint
from minFQ.run_data_tracker import RunDataTracker

log = logging.getLogger(__name__)

class OpenLine:
    def __init__(self, fp, start=1, number=-1, f=open, f_kwds=None):
        """

        Generic function to return a generator for reading a file from a given line.

        Parameters
        ----------
        fp : str
            Filepath to the file.
        start : int
            Starting line number. Default 1.
        number : int
            The number of lines to read. If -1, will read to EOF. Default -1
        f : func
            The opening function that is used on the provided filepath. Default open
        f_kwds: dict
            The keyword args to pass to the open function. Default None.
        """
        self.fp = fp
        self.start = start
        self.number = number
        self.open_func = f
        self.current_line = 0

        if number == -1:
            self.number = float("inf")

        if f_kwds is None:
            self.f_kwds = {}
        else:
            self.f_kwds = f_kwds

    def __enter__(self):
        with self.open_func(self.fp, **self.f_kwds) as fh:
            for i, L in enumerate(fh, start=1):
                if i < self.start:
                    continue
                if i >= self.start + self.number:
                    break
                self.current_line = i
                yield L.strip()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True


def get_file_size(file_path):
    """
    Returns the size of the filepath in bytes.
    :param filePath: The path to file in the watch directory
    :return:
    """
    return os.path.getsize(file_path)


def split_all(path):
    """
    Split the path into it's relative parts so we can check all the way down the tree
    :param path: Path provided by the user to the watch directory
    :return: A list of paths created from the base path, so we can use relative and absolute paths
    """
    all_parts = []
    while 1:
        parts = os.path.split(path)

        if parts[0] == path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts


def check_fastq_path(path):
    """
    Check if the files we are looking at are directly inside pass or fail folders, or one or two removed.
    :param path:
    :return:
    """
    folders_in_path = split_all(path)
    try:
        if folders_in_path[-2] in ("pass", "fail"):
            return "{}_{}".format(folders_in_path[-2], folders_in_path[-1])
        elif folders_in_path[-3] in ("pass", "fail"):
            return "{}_{}_{}".format(folders_in_path[-3], folders_in_path[-2], folders_in_path[-1])
        else:
            return "{}".format(folders_in_path[-1])
    except Exception as e:
        return "{}".format(folders_in_path[-1])


def check_is_pass(path, avg_quality):
    """
    Check that the fastq file is contains pass or fail reads
    Parameters
    ----------
    path
    avg_quality

    Returns
    -------

    """
    folders = os.path.split(path)
    if "pass" in folders[0]:
        return True
    elif "fail" in folders[0]:
        return False
    else:
        if avg_quality != None:
            if avg_quality >= 7:
                return True  # This assumes we have been unable to find either pass
                # or fail and thus we assume the run is a pass run.
            else:
                return False
        else:
            return True


def parse_fastq_description(description):
    """
    Parse the description found in a fastq reads header

    Parameters
    ----------
    description: str
        A string of the fastq reads description header

    Returns
    -------
    description_dict: dict
        A dictionary containing the keys and values found in the fastq read headers
    """
    description_dict = {}
    descriptors = description.split(" ")
    # Delete symbol for header
    for item in descriptors:
        if "=" in item:
            bits = item.split("=")
            description_dict[bits[0]] = bits[1]
    return description_dict


def _prepare_toml(toml_dict):
    """
    Prepares the dictionary of the toml, places the channel number as key and conditon name as value
    Returns
    -------
    dict
    """
    _d = {}
    # Reverse the dict so we can lookup name by channel
    for key in toml_dict["conditions"].keys():
        channels = toml_dict["conditions"][key]["channels"]
        name = toml_dict["conditions"][key]["name"]
        _d.update({channel: name for channel in channels})
    return _d


def average_quality(quality):
    """
    Return the average quality of a read from it's quality string
    Parameters
    ----------
    quality: str

    Returns
    -------
    float
    """
    return -10 * np.log10(
        np.mean(
            10
            ** (
                np.array(
                    (-(np.array(list(quality.encode("utf8"))) - 33)) / 10, dtype=float
                )
            )
        )
    )


def get_runid(fastq):
    """
    Open a fastq file, read the first line and parse out the Run ID
    :param fastq: path to the fastq file to be parsed
    :type fastq: str
    :return runid: The run ID of this fastq file as a string
    """
    runid = ""
    if fastq.endswith(".gz"):
        with gzip.open(fastq, "rt") as file:
            for _ in range(1):
                line = file.readline()
    else:
        with open(fastq, "r") as file:
            for _ in range(1):
                line = file.readline()
    for _ in line.split():
        if _.startswith("runid"):
            runid = _.split("=")[1]
    return runid


def unseen_files_in_watch_folder_dict(path, ignore_existing, minotour_api, fastq_dict, sequencing_statistics):
    """
    Iterate fastq files in watch directory and see if we have seen them before by checking against server.
    Parameters
    ----------
    path: str
        File path to watch directory
    ignore_existing: bool
        Ignore existing fastq folders in the direcotur
    minotour_api: minFQ.minotourapi.MinotourAPI
        The minotour api
    fastq_dict: dict
        Dictionary containing all the fastq files that we have seen.
    sequencing_statistics: minFQ.utils.SequencingStatistics
        Tracker for files in watch list and uplaod metrics

    Returns
    -------
    dict
        Dictionary of fastq files that we haven't seen before. Key filepath, value created time for fastq file

    """
    # Dictionary for tracking files
    new_fastq_file_dict = {}
    store_path = path
    if not ignore_existing:
        counter = 0
        # if directory
        if os.path.isdir(path):
            log.info("caching existing fastq files in: {}".format(path))
            sequencing_statistics.fastq_message = "caching existing fastq files in: {}".format(path)
            ## ToDo Consider moving these to top level
            novel_run_set = set()
            seen_file_tracker = {}
            file_endings = {".fq", ".fastq", ".fq.gz", ".fastq.gz", ".fq.xz", ".fastq.xz"}
            # have s rummage around the watch directory
            for path, dirs, files in os.walk(path):
                # iterate fastq files in the watchdir
                for f in files:
                    if "".join(Path(f).suffixes) in file_endings:
                        log.debug("Processing File {}\r".format(f))
                        counter += 1
                        sequencing_statistics.files_seen += 1
                        #### Here is where we want to check if the files have been created and what the checksums are
                        #### If the file checksums do not match, we pass the file to the rest of the script.
                        #### When we finish analysing a file, we will need to update this information n the server.
                        #### Currently just using size.
                        file_byte_size = get_file_size(os.path.join(path, f))
                        run_id = get_runid(os.path.join(path, f))
                        sequencing_statistics.fastq_info[run_id]["files_seen"] += 1
                        sequencing_statistics.fastq_info[run_id]["run_id"] = run_id
                        if "directory" not in sequencing_statistics.fastq_info[run_id]:
                            sequencing_statistics.fastq_info[run_id]["directory"] = store_path
                        if (
                            run_id not in novel_run_set
                            and run_id not in seen_file_tracker.keys()
                        ):
                            # get all files for this run
                            result = minotour_api.get_json(EndPoint.FASTQ_FILE, base_id=run_id)
                            #### Here we want to parse through the results and store them in some kind of dictionary in order that we can check what is happening
                            # We are parsing through the fastq files we have seen for this run so we don't reprocess them
                            if result is not None:
                                for entry in result:
                                    if entry["runid"] not in seen_file_tracker.keys():
                                        seen_file_tracker[entry["runid"]] = {}
                                    seen_file_tracker[entry["runid"]][
                                        entry["name"]
                                    ] = entry["md5"]
                        # add the file path
                        filepath = os.path.join(path, f)
                        check_file_path = check_fastq_path(filepath)
                        if check_file_path not in fastq_dict.keys():
                            fastq_dict[check_file_path] = {}
                        fastq_dict[check_file_path]["runid"] = run_id
                        fastq_dict[check_file_path]["md5"] = file_byte_size
                        """Here we are going to check if the files match or not. """
                        seen_file = False
                        if (
                            run_id in seen_file_tracker.keys()
                            and check_file_path in seen_file_tracker[run_id]
                        ):
                            seen_file = True
                            ## if the file hasn't changed size, we can skip it
                            if int(file_byte_size) == int(
                                seen_file_tracker[run_id][check_file_path]
                            ):
                                sequencing_statistics.fastq_info[run_id]["files_skipped"] += 1
                                sequencing_statistics.files_skipped += 1
                            else:
                                new_fastq_file_dict[filepath] = os.stat(filepath).st_mtime
                                novel_run_set.add(run_id)

                        if not seen_file:
                            # never seen this file before
                            new_fastq_file_dict[filepath] = os.stat(filepath).st_mtime
                            novel_run_set.add(run_id)
        log.info("processed {} files".format(counter))
        sequencing_statistics.fastq_message = "processed {} files".format(counter)
        log.info(
            "found {} existing fastq files to process first.".format(
                len(new_fastq_file_dict)
            )
        )
    else:
        sequencing_statistics.fastq_message = "Ignoring existing fastq files in: {}".format(path)
    return new_fastq_file_dict


def create_run_collection(run_id, run_dict, args, header, description_dict, sequencing_statistics):
    """
    Create run collection for this run id if we don't already have one, store in run_dict
    Parameters
    ----------
    run_id: str
        The uuid of this run
    run_dict: dict
        The dictionary containing run collections
    args: argparse.NameSpace
        The command line arguments
    header: dict
        Authenticator for request
    description_dict: dict
        The description dict
    sequencing_statistics: SequencingStatistics
        Class to communicate sequencing statistics
    Returns
    -------

    """
    run_dict[run_id] = RunDataTracker(args, header, sequencing_statistics)
    run_dict[run_id].add_run(description_dict, args)




class MTConnectionError(Exception):
    """
    Exception raised when we get an unexpected status code from a connection to the minoTour sever
    """

    def __init__(
        self, response, message="Unexpected status code (not in expected values)"
    ):
        """

        Parameters
        ----------
        response: requests.models.Response
            The requests object from the request
        message: str
            The message to display at the bottom of the Exception
        """
        self.response = response
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        """
        Overides the displayed message string
        Returns
        -------
        str
        """
        lookup_message = {401: "Nuh uh - unauthorised", 500: "MinoTour says no"}
        return "{} \n -> {} \n -> \n {} \n -> {}".format(
            self.response.status_code,
            lookup_message.get(self.response.status_code, ""),
            self.message,
            self.response.text,
        )


def except_rpc(wrapped_function):
    def _wrapper(*args, **kwargs):
        try:
            # do something before the function call
            result = wrapped_function(*args, **kwargs)
            # do something after the function call
        except TypeError:
            print("TypeError")
        except IndexError:
            print("IndexError")
        # return result

    return _wrapper



from enum import Enum


class EndPoint(Enum):
    """
    An Enum to store all the endpoints in a single place, so we can track them easily.
    Import to the files you need endpoints in and add them to minotourapi requests.

    """
    # Get
    REFERENCES = "/reference/"
    VERSION = "/minknow/version/"
    GET_MINION = "/minknow/minions/{}/"
    TASK_TYPES = "/reads/task-types/"
    TARGET_SETS = "/metagenomics/target-sets"
    PRIMER_SCHEMES = "/artic/schemes"
    # todo if we get the api right, we can just combine the base minion with the status or events etc.
    MINION_STATUS = "/minknow/minions/{}/status/"
    TEST = "/minknow/test-connect/"
    READ_TYPES = "/reads/read-types/"
    READ_NAMES = "/reads/runs/{}/read-names/"
    MINION_CONTROL = "/minknow/minions/{}/control/"

    # Post or Put
    MINION_EVENT_TYPES = "/minknow/minions/event-types/"
    MINION_EVENT = "/minknow/minions/{}/events/"
    MINION_MESSAGES = "/minknow/minions/{}/messages/"
    MINION_RUN_INFO = "/minknow/minions/runs/{}/"
    MINION_RUN_STATS = "/minknow/minions/runs/{}/run-stats/"
    POST_MINION = "/minknow/minions/"
    RUNS = "/reads/runs/{}/"
    FLOWCELL = "/reads/flowcells/{}/"
    FASTQ_FILE = "/reads/runs/{}/files/"
    BARCODE = "/reads/barcode/"
    JOBS = "/reads/tasks/"
    READS = "/reads/read/"

    def __str__(self):
        return "{}".format(self.value)

    def add_id(self, id):
        """
        Return the partial URL, with id formatted into it if necessary
        Parameters
        ----------
        id: str
            The primary key of the minion that we want to use

        Returns
        -------

        """
        return self.value.format(id)

    def append_id(self, base_id, append_id):
        """
        Append ID to the end of a URL
        Parameters
        ----------
        base_id: str
            base id
        append_id: str
            Appended ID

        Returns
        -------

        """
        return self.add_id(base_id) + str(append_id) + "/"

    def strip_id(self):
        """
        Strip the id format space from the URL, so we can use it for posts
        Returns
        -------

        """
        return self.value.split("{}")[0]

    def resolve_url(self, base_id="", append_id="", no_id=False):
        """
        Resolve the url in regards to id in the str
        Parameters
        ----------
        base_id: str
            The base id for the url ex. /minion/*1*/
        append_id: str
            The appended id after the base id ex. /minion/1/jobs/*2*
        no_id: bool
            If True Strip the id space from the url

        Returns
        -------
        str
            The Url to append to the request
        """
        if base_id and not append_id:
            return self.add_id(base_id)
        elif base_id and append_id:
            return self.append_id(base_id, append_id)
        elif no_id:
            return self.strip_id()
        else:
            return str(self)



"""
File Routines for handling fastq files and monitoring locations. Built on watchdog.
"""
import gzip
import logging
import lzma
import sys
import threading
import time
import toml as toml_manager

from minFQ.endpoints import EndPoint
from minFQ.fastq_handler_utils import (
    parse_fastq_description,
    get_file_size,
    get_runid,
    check_fastq_path,
    unseen_files_in_watch_folder_dict,
    _prepare_toml,
    OpenLine,
    create_run_collection, average_quality, check_is_pass,
)
from watchdog.events import FileSystemEventHandler

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


# ##Function modified from https://raw.githubusercontent.com/lh3/readfq/master/readfq.py

def readfq(fp):  # this is a generator function
    last = None  # this is a buffer keeping the last unprocessed line
    while True:  # mimic closure; is it a bad idea?
        if not last:  # the first record or a record following a fastq
            for l in fp:  # search for the start of the next record
                if l[0] in ">@":  # fasta/q header line
                    last = l[:-1]  # save this line
                    break
        if not last:
            break
        desc, name, seqs, last = last[1:], last[1:].partition(" ")[0], [], None
        for l in fp:  # read the sequence
            if l[0] in "@+>":
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != "+":  # this is a fasta record
            yield desc, name, "".join(seqs), None  # yield a fasta record
            if not last:
                break
        else:  # this is a fastq record
            seq, leng, seqs = "".join(seqs), 0, []
            for l in fp:  # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq):  # have read enough quality
                    last = None
                    yield desc, name, seq, "".join(seqs)  # yield a fastq record
                    break
            if last:  # reach EOF before reading enough quality
                yield desc, name, seq, None  # yield a fasta record instead
                break


def parse_fastq_record(
    name,
    seq,
    qual,
    fastq_file_path,
    run_dict,
    args,
    fastq_file,
    counter,
    sequencing_statistic,
    description_dict,
):
    """

    Parameters
    ----------
    name: str
        The fastq read id
    seq: str
        The sequence string for this fastq read
    qual: str
        The quality string for this fastq read
    fastq_file_path: str
        The absolute path to this fastq file
    run_dict: dict
        Dictionary containg runs that we have seen before and their RunCollection
    args: argparse.NameSpace
        The command line argumanets chosen
    fastq_file: dict
        Dictionary with the minoTour record info for the fastq file this read comes from
    counter: int
        The counter for the number of reads taht we have seen
    sequencing_statistic: SequencingStatistics
        The sequencing statistics class to track metrics about the run
    description_dict: dict
        The description dictionary, split by key to value

    Returns
    -------

    """
    fastq_read = {
        "read": description_dict.get("read", None),
        "run_id": description_dict.get("runid", None),
        "channel": description_dict.get("ch", None),
        "start_time": description_dict.get("start_time", None),
        "read_id": name,
        "sequence_length": len(seq),
        "fastq_file": fastq_file["id"],
        "quality": qual,
        "sequence": seq
    }
    if args.skip_sequence:
        fastq_read["sequence"] = ""
        fastq_read["quality"] = ""
    if args.toml is not None:
        try:
            toml_dict = toml_manager.load(args.toml)
            toml_dict = _prepare_toml(toml_dict)
        except FileNotFoundError as e:
            sys.exit(
                "Error, toml file not found. Please check that it hasn't been moved."
            )
        run_dict[fastq_read["run_id"]].add_toml_file(toml_dict)
    if args.unblocks is not None:
        run_dict[fastq_read["run_id"]].unblocked_file = args.unblocks
    ### Check the overlap between the current file path and the folders being watched:
    for folder in sequencing_statistic.watched_directory_set:
        if folder is not None:
            if fastq_file_path.startswith(folder):
                if run_dict[fastq_read["run_id"]].run_folder != folder:
                    ## We have found the folder that this fastq file comes from.
                    run_dict[fastq_read["run_id"]].add_run_folder(folder)


    if counter <= 1:
        ## This is the first read we have seen from this file - so we are going to check for updates in the unblocked read file.
        if run_dict[fastq_read["run_id"]].unblocked_file is not None:
            with OpenLine(
                run_dict[fastq_read["run_id"]].unblocked_file,
                run_dict[fastq_read["run_id"]].unblocked_line_start,
            ) as fh:
                _d = {line: 1 for line in fh}
                lines_returned = len(_d)
                run_dict[fastq_read["run_id"]].unblocked_dict.update(_d)
            run_dict[fastq_read["run_id"]].unblocked_line_start += lines_returned

    if fastq_read["read_id"] not in run_dict[fastq_read["run_id"]].read_names:
        quality = qual
        # Turns out this is not the way to calculate quality...
        # This is slow.
        if quality is not None:
            fastq_read["quality_average"] = round(average_quality(quality), 2,)
        fastq_read["is_pass"] = check_is_pass(fastq_file_path, fastq_read["quality_average"])
        # use 'No barcode' for non-barcoded reads
        barcode_name = description_dict.get("barcode", None)
        fastq_read["barcode_name"] = barcode_name.replace(" ", "_") if barcode_name else "No_barcode"
        # Parse the channel out of the description and lookup it's corresponding condition
        # set it to the reads barcode
        if run_dict[fastq_read["run_id"]].toml is not None and args.concat_condish:
            fastq_read["barcode_name"] = fastq_read["barcode_name"] + "_" + run_dict[fastq_read["run_id"]].toml[
                int(fastq_read["channel"])
            ]
        is_unblocked = fastq_read["read_id"] in run_dict[fastq_read["run_id"]].unblocked_dict
        fastq_read["rejected_barcode_name"] = "Unblocked" if is_unblocked else "Sequenced"
        # add control-treatment if passed as argument
        if args.treatment_control:
            is_control = int(fastq_read["channel"]) % args.treatment_control == 0
            suffix = " - control" if is_control else " - treatment"
            fastq_read["barcode_name"] = fastq_read["barcode_name"] + suffix
        # Add the read to the class dictionary to be uploaded, pretty important
        run_dict[fastq_read["run_id"]].add_read(fastq_read)
    else:
        sequencing_statistic.reads_skipped += 1
        sequencing_statistic.fastq_info[fastq_read["run_id"]]["reads_skipped"] += 1


def parse_fastq_file(
    fastq_path, run_dict, args, header, minotour_api, sequencing_stats
):
    """
    
    Parameters
    ----------
    fastq_path: str
        Path to the fastq file to parse
    run_dict: dict
        Dictionary containing the run
    args: argparse.NameSpace
        The namespace of the parsed command line args
    header: dict
        The authentication header to make requests to the server
    minotour_api: minFQ.minotourapi.MinotourAPI
        The minotour API class instance for convenience request to the server
    sequencing_stats: minFQ.utils.SequencingStatistics
        The class to track the sequencing upload metrics and files

    Returns
    -------
    int 
        The updated number of lines we have already seen from the unblocked read ids file
    """
    # The updated number of lines we have already seen from the unblocked read ids file
    log.debug("Parsing fastq file {}".format(fastq_path))
    # Get runId from the path
    run_id = get_runid(fastq_path)
    payload = {
        "file_name": str(check_fastq_path(fastq_path)),
        "run_id": run_id,
        "md5": "0",
        "run": None
    }
    if run_id in run_dict:
        fastq_file = minotour_api.post(
            EndPoint.FASTQ_FILE, json=payload, base_id=run_id
        )
    if not sequencing_stats.fastq_info[run_id]["run_id"]:
        sequencing_stats.fastq_info[run_id]["run_id"] = run_id
    counter = 0
    # fq = pyfastx.Fastq(fastq_path)
    # gen = (read for read in fq)
    # for read in gen:
    open_handle_dict = {
        ".gz": gzip.open,
        ".xz": lzma.open
    }
    handle = open_handle_dict.get(fastq_path[-3:], open)
    with handle(fastq_path, "rt") as fh:
        for desc, name, seq, qual in readfq(fh):
            description_dict = parse_fastq_description(desc)
            counter += 1
            sequencing_stats.reads_seen += 1
            sequencing_stats.fastq_info[run_id]["reads_seen"] += 1
            sequencing_stats.fastq_message = "processing read {}".format(counter)
            try:
                ### We haven't seen this run before - so we need to check stuff.
                # create runCOllection in the runDict
                if description_dict["runid"] not in run_dict:
                    create_run_collection(
                        run_id,
                        run_dict,
                        args,
                        header,
                        description_dict,
                        sequencing_stats,
                    )
                    fastq_file = minotour_api.post(
                        EndPoint.FASTQ_FILE, json=payload, base_id=run_id
                    )
                    run_dict[run_id].get_readnames_by_run(fastq_file["id"])
                parse_fastq_record(
                    name,
                    seq,
                    qual,
                    fastq_path,
                    run_dict,
                    args,
                    fastq_file,
                    counter=counter,
                    sequencing_statistic=sequencing_stats,
                    description_dict=description_dict,
                )

            except Exception as e:
                sequencing_stats.reads_corrupt += 1
                log.error(e, exc_info=True)
                log.error("This gzipped file failed to upload - {}.".format(fastq_path))
                raise Exception
                # continue
    # This chunk of code will mean we commit reads every time we get a new file?
    for runs in run_dict:
        run_dict[runs].read_names = []
        run_dict[runs].commit_reads()
    try:
        # Update the fastq file entry on the server that says we provides file size to database record
        payload = {
            "file_name": str(check_fastq_path(fastq_path)),
            "run_id": run_id,
            "md5": get_file_size(fastq_path),
            "run": run_dict[run_id].run,
        }
        fastq_file = minotour_api.put(EndPoint.FASTQ_FILE, json=payload, base_id=run_id)
    except Exception as err:
        log.error("Problem with uploading file {}".format(err))
    sequencing_stats.time_per_file = time.time()
    sequencing_stats.fastq_info[run_id]["files_processed"] += 1
    return counter


class FastqHandler(FileSystemEventHandler):
    def __init__(self, args, header, rundict, sequencing_statistic, minotour_api):
        """
        Collect information about fastq files that have been written out.
        Parameters
        ----------
        args: argparse.NameSpace
            The arguments provided on the command line
        header: dict
            The authentication header for the requests
        rundict: dict
            The dictionary for tracking the runs
        sequencing_statistic: minFQ.utils.SequencingStatistics
            The class for tracking files and metrics about the upload
        minotour_api: minFQ.minotourapi.MinotourAPI
            The API class for convenience connecting to MinoTour server
        """
        self.sequencing_statistic = sequencing_statistic
        self.file_descriptor = {}
        self.args = args
        self.header = header
        # adding files to the file_descriptor is really slow - therefore lets skip that and only update the files when we want to basecall thread_number
        self.minotour_api = minotour_api
        self.rundict = rundict
        # Dict of all fastq files that we have seen, irrespective of if they exists on server or not
        self.fastq_dict = {}
        self.toml_dict = {}
        self.fastq_files_to_create = {}
        self.processing = {}
        self.running = True
        self.t = threading.Thread(target=self.process_files)
        try:
            self.t.start()
        except Exception as e:
            self.running = False
            self.args.errored = True
            self.args.error_message = "Flowcell name is required. This may be old FASTQ data, please provide a name with -n."
            log.error(
                "Flowcell name is required. This may be old FASTQ data, please provide a name with -n."
            )

    def addfolder(self, folder):
        self.fastq_files_to_create.update(
            unseen_files_in_watch_folder_dict(
                folder,
                self.args.ignore_existing,
                self.minotour_api,
                self.fastq_dict,
                self.sequencing_statistic,
            )
        )

    def stopt(self):
        self.running = False

    def lencreates(self):
        return len(self.fastq_files_to_create)

    def process_files(self):
        """
        Process fastq files in a thread. This is the work horse for read fastqs. We have one of these per watch directory in the watch list
        :return:
        """
        while self.running:
            current_time = time.time()
            # sort by created time descending
            for fastqfile, create_time in sorted(
                self.fastq_files_to_create.items(), key=lambda x: x[1]
            ):
                # remove the file from fastq files to create as we are doing that now
                if create_time > time.time() - 5:
                    time.sleep(5)
                # file created 5 sec ago, so should be complete. For simulations we make the time longer.
                del self.fastq_files_to_create[fastqfile]
                parse_fastq_file(
                    fastqfile,
                    self.rundict,
                    self.args,
                    self.header,
                    self.minotour_api,
                    self.sequencing_statistic,
                )
                self.sequencing_statistic.files_processed += 1

                if not self.running:
                    break
            if current_time + 5 > time.time():
                time.sleep(5)

    def on_created(self, event):
        """Watchdog counts a new file in a folder it is watching as a new file"""
        """This will add a file which is added to the watchfolder to the creates and the info file."""
        # if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz")):
        #     self.creates[event.src_path] = time.time()

        log.debug("Processing created file {}".format(event.src_path))
        # time.sleep(5)
        if (
            event.src_path.endswith(".fastq")
            or event.src_path.endswith(".fastq.gz")
            or event.src_path.endswith(".fq")
            or event.src_path.endswith(".fq.gz")
        ):
            self.sequencing_statistic.files_seen += 1
            run_id = get_runid(event.src_path)
            self.sequencing_statistic.fastq_info[run_id]["files_seen"] += 1
            self.fastq_files_to_create[event.src_path] = time.time()

    def on_modified(self, event):
        if (
            event.src_path.endswith(".fastq")
            or event.src_path.endswith(".fastq.gz")
            or event.src_path.endswith(".fq")
            or event.src_path.endswith(".fq.gz")
        ):
            log.debug("Modified file {}".format(event.src_path))
            self.fastq_files_to_create[event.src_path] = time.time()

    def on_moved(self, event):
        if any(
            (
                event.dest_path.endswith(".fastq"),
                event.dest_path.endswith(".fastq.gz"),
                event.dest_path.endswith(".fq"),
                event.dest_path.endswith(".fq.gz"),
            )
        ):
            log.debug("Modified file {}".format(event.dest_path))
            self.fastq_files_to_create[event.dest_path] = time.time()



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))
# Get version
version = {}
with open("minFQ/version.py") as fp:
    exec(fp.read(), version)
setup(
    name="minFQ",
    version=version["__version__"],
    description="Command line interface for uploading fastq files and monitoring minKNOW for minotour.",
    long_description=open(path.join(here, "README.rst")).read(),
    url="https://github.com/LooseLab/minotourcli",
    author="Matt Loose",
    author_email="matt.loose@nottingham.ac.uk",
    license="MIT",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="nanopore quality control analysis",
    packages=find_packages(),
    python_requires=">=3.5",
    setup_requires=["numpy"],
    install_requires=[
        "tqdm",
        "python-dateutil",
        "requests",
        "numpy",
        "watchdog",
        "configargparse",
        "grpcio",
        "google",
        "protobuf",
        "pandas",
        "ont-fast5-api",
        "minknow-api",
        "validators",
        "grpcio-tools",
        "toml",
    ],
    package_data={"minFQ": []},
    package_dir={"minFQ": "minFQ"},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "minFQ=minFQ.minFQ:main",
        ],
    },
)



import os
import sys
import platform
import logging
import time
import curses
from logging.handlers import RotatingFileHandler

from minFQ.minknow_connection import MinionManager
from minFQ.fastq_handler import FastqHandler
import configargparse
from watchdog.observers.polling import PollingObserver as Observer
from minFQ.minotourapi import MinotourAPI
from minFQ.utils import (
    add_arguments_to_parser,
    validate_args,
    check_server_compatibility,
    SequencingStatistics,
    list_minotour_options,
    check_job_from_client,
    write_out_minfq_info,
    write_out_minknow_info,
    write_out_fastq_info,
    refresh_pad,
    ascii_minotour,
    CursesHandler,
)

logging.basicConfig(
    handlers=[RotatingFileHandler("./minFQ.log", maxBytes=20000000, backupCount=3)],
    level=logging.INFO,
    format="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)

log = logging.getLogger()
log.setLevel(logging.DEBUG)

# define a Handler which writes INFO messages or higher to the sys.stderr
# log = logging.getLogger("minFQ")


def start_minknow_and_basecalled_monitoring(
    sequencing_statistics, args, header, minotour_api, stdscr, screen, log_win
):
    """
    Start the minKnow monitoring and basecalled data monitoring in accordance with arguments passed by user
    Parameters
    ----------
    sequencing_statistics: minFQ.utils.SequencingStatistics
        Tracker class for files being monitored, and the metrics about upload
    args: argparse.NameSpace
        The command line arguments that were passed to the script
    log: logging.Logger
        The logger for this script
    header: dict
        The dictionary with headers for the requests, including authentiction
    minotour_api: minFQ.minotourapi.MinotourAPI
        The minotourAPI class
    stdscr: _curses.window
        Curses window for printing out
    screen: _curses.window
        The main curses screen
    log_win: _curses.window
        The logging window we write to
    Returns
    -------

    """
    sequencing_statistics.read_count = 0
    runs_being_monitored_dict = {}
    if not args.no_fastq:
        # This block handles the fastq
        # Add our watchdir to our WATCHLIST
        if args.watch_dir is not None:
            sequencing_statistics.to_watch_directory_list.append(args.watch_dir)
        # if we are connecting to minKNOW
    if not args.no_minknow:
        # this block is going to handle the running of minknow monitoring by the client.
        stdscr.addstr("Connecting to minknow instance at {}".format(args.ip))
        refresh_pad(screen, stdscr)
        minknow_connection = MinionManager(
            args=args, header=header, sequencing_statistics=sequencing_statistics
        )
    curses.napms(2000)
    stdscr.clear()
    event_handler = FastqHandler(
        args, header, runs_being_monitored_dict, sequencing_statistics, minotour_api
    )
    observer = Observer()
    observer.start()
    stats = True
    try:
        while True:
            try:
                c = stdscr.getch()
                if c == ord("l"):
                    stats = False
                elif c == ord("s"):
                    stats = True
            except curses.ERR:
                stats = True
            # todo these should be abstracted into one function as they are verrrry similar
            if stats:
                stdscr.addstr(
                    0,
                    0,
                    "To stop minFQ use CTRL-C. To see the logs, Press l. To return to stats, Press s.",
                    curses.color_pair(4),
                )
                stdscr.addstr(1, 0, "Fetching Data...")
                write_out_minfq_info(stdscr, sequencing_statistics)
                ascii_minotour(stdscr)
                write_out_minknow_info(stdscr, sequencing_statistics)
                write_out_fastq_info(stdscr, sequencing_statistics)
                refresh_pad(screen, stdscr)
                stdscr.overwrite(screen)
            else:
                log_win.overwrite(screen)
                # log_win.addstr(0, 0, "To stop minFQ use CTRL-C. To see the logs, Press l. To see info, Press s.", curses.color_pair(4))
                refresh_pad(screen, log_win)
            screen.refresh()
            if not args.no_fastq and sequencing_statistics.to_watch_directory_list:
                for (
                    folder
                ) in (
                    sequencing_statistics.to_watch_directory_list
                ):  # directory watchlist has the new run dir
                    log.warning(
                        "Checking folder {} that is in our to watch directory".format(
                            folder
                        )
                    )
                    if (
                        folder
                        and folder not in sequencing_statistics.watched_directory_set
                    ):
                        # check that the folder exists, before adding it to be scheduled
                        if os.path.exists(folder):
                            # We have a new folder that hasn't been added.
                            # We need to add this to our list to schedule and catalogue the files.
                            # TODO bug here where we never remove directories from the already watching set - eventually this would lead to a massive set of strings in this set if minFQ is never quit
                            sequencing_statistics.watched_directory_set.add(folder)
                            event_handler.addfolder(folder)
                            log.info("FastQ Monitoring added for {}".format(folder))
                            sequencing_statistics.to_watch_directory_list.remove(folder)
                            sequencing_statistics.update = True
                        else:
                            log.warning(
                                "Waiting for minKNOW to create folder {} before updating watchdog.".format(
                                    folder
                                )
                            )
                log.warning(
                    "Updating observers is {}".format(sequencing_statistics.update)
                )
                if sequencing_statistics.update:
                    observer.unschedule_all()
                    for folder in sequencing_statistics.watched_directory_set:
                        if folder and os.path.exists(folder):
                            log.info(
                                "Scheduling observer for {}, which does exist".format(
                                    folder
                                )
                            )
                            observer.schedule(
                                event_handler, path=folder, recursive=True
                            )
                        else:
                            log.warning(
                                "Tried to add {}, but folder does not exist".format(
                                    folder
                                )
                            )
                    sequencing_statistics.update = False
                # check if we need to remove any fastq info
            sequencing_statistics.check_fastq_info()
            time.sleep(1)

            if sequencing_statistics.errored:
                log.error("Errored - Will take a few seconds to clean clear_lines!")
                log.error(sequencing_statistics.error_message)
                if not args.no_minknow:
                    minknow_connection.stop_monitoring()
                if not args.no_fastq:
                    observer.stop()
                    observer.join()
                observer.stop()
                observer.join()
                event_handler.stopt()
                curses.nocbreak()
                stdscr.keypad(False)
                curses.echo()
                curses.endwin()
                print(repr(sequencing_statistics.error_message))
                print("Exiting after error - Will take a few seconds to close threads.")
                sys.exit(0)

    except (KeyboardInterrupt, Exception) as e:
        stdscr.addstr("Exiting - Will take a few seconds to close threads.")
        if not args.no_minknow:
            minknow_connection.stop_monitoring()
        observer.stop()
        observer.join()
        event_handler.stopt()
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()
        curses.endwin()
        print(repr(e))
        print("Exiting - Will take a few seconds to close threads.")
        sys.exit(0)


def main():
    """
    Entry point for minFQ, parses CLI arguments and sets off correct scripts
    Returns
    -------

    """
    screen = curses.initscr()
    num_rows, num_cols = screen.getmaxyx()
    stdscr = curses.newpad(200, 200)
    stdscr.scrollok(True)
    stdscr.nodelay(True)
    curses.noecho()
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_GREEN, -1)
    curses.init_pair(2, curses.COLOR_RED, -1)
    curses.init_pair(3, curses.COLOR_BLUE, -1)
    curses.init_pair(4, curses.COLOR_MAGENTA, -1)
    curses.init_pair(5, curses.COLOR_YELLOW, -1)
    stdscr.clear()
    ## create log window
    num_rows, num_cols = screen.getmaxyx()
    log_win = curses.newpad(60, 200)
    log_win.scrollok(True)
    log_win.nodelay(True)
    log_win.idlok(True)
    log_win.leaveok(True)

    mh = CursesHandler(screen, log_win)
    mh.setLevel(logging.DEBUG)
    formatterDisplay = logging.Formatter(
        "%(asctime)s %(module)s:%(levelname)s:%(thread)d:%(message)s"
    )
    mh.setFormatter(formatterDisplay)
    log.addHandler(mh)
    if platform.system() == "Windows":  # MS
        config_file = os.path.join(os.path.sep, sys.prefix, "minfq_windows.config")
    else:
        config_file = os.path.join(
            os.path.sep,
            os.path.dirname(os.path.realpath("__file__")),
            "minfq_unix.config",
        )

    parser = configargparse.ArgParser(
        description="minFQ: A program to analyse minION fastq files in real-time or post-run and monitor the activity of MinKNOW.",
        default_config_files=[config_file],
    )
    parser = add_arguments_to_parser(parser, stdscr)
    # Get the arguments namespace
    args = parser.parse_args()
    handler = logging.getLogger().handlers[0]
    # old_level = console_handler.level
    # console_handler.setLevel(logging.DEBUG)
    handler.setLevel(getattr(logging, args.loglevel.upper()))
    log.setLevel(getattr(logging, args.loglevel.upper()))
    # log = configure_logging(getattr(logging, args.loglevel.upper()), screen, log_win)
    stdscr.addstr(
        str(
            """Welcome to
           _      ______ _____ 
          (_)     |  ___|  _  |
 _ __ ___  _ _ __ | |_  | | | |
| '_ ` _ \| | '_ \|  _| | | | |
| | | | | | | | | | |   \ \/' /
|_| |_| |_|_|_| |_\_|    \_/\_\ \n"""
        ),
        curses.color_pair(2),
    )
    refresh_pad(screen, stdscr)
    header = {
        "Authorization": "Token {}".format(args.api_key),
        "Content-Type": "application/json",
        "Content-Encoding": "gzip",
    }
    stdscr.addstr(
        """MMMMMMMMMMMNNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMWWWMMMMMMMMMMM
MMMMMMMMMMWkkNMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMWKOXMMMMMMMMMMM
MMMMMMMMMMWx:dKNWMMMMMMMMMMMMMMMMMMMMMMMMMMMMMWWXkcoXMMMMMMMMMMM
MMMMMMMMMMMKl;:lxkO0KKKXXNNNNWWWWWNNNNXXXKKK0Oxoc;:kWMMMMMMMMMMM
MMMMMMMMMMMWKd:;,,;;;:::ccclllllllllcccc:::;;,,,;lONMMMMMMMMMMMM
MMMMMMMMMMMMMWXOxollcc::;;,,,,,,,,,,,,;:::cclodk0NMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMWWXK00KOo;,,,,,,,,,,:d0KK0KNWWMMMMMMMMMMMMMMMMM
MMMMMMMMMNKkxoc:;;,,''xW0c,,,,,,,,,,,,lKWx'',,;;:loxOKWMMMMMMMMM
MMMMWXko;'.          .dWk;,,,,,,,,,,,,:kWd           .':oONMMMMM
MMNkc.                oWO:,,,,,,,,,,,,c0Wo                .cONMM
WO;                   :XXd;,,,,,,,,,,;dNK;                   ;0W
x.                     lNKo;,,,,,,,,;oXXc                     .k
.                       lXXxc,,,,,;cxXXc                       '
            .co;.        'xXKOxddxOXXd'        .:oc.            
           .xWMW0o'        'lxkOOkxl'        'oKWMWd.          .
'          ;XMMMMMXl.                      .oXMMMMMK,          ,
k.         ,0MMMMMMWO,                    ,OWMMMMMMO'         .O
WO'        .dWMMMMMMMX:                  cXMMMMMMMWo         'OM
MMKl.       'OKdc:lkNMK;                :XMXxl:cdKk.       .lXMM
MMMW0c.      ..     ;KM0,              ;KM0,     ..      .c0WMMM
MMMMMWKo,.           oMMd             .xMWl           .,dKWMMMMM
MMMMMMMMW0o;.        oMM0'            ,KMWl        .;d0WMMMMMMMM
MMMMMMMMMMMWXko:,. .cKMMWl            lWMMK:  .,:oOXWMMMMMMMMMMM
MMMMMMMMMMMMMMMMWXO0NWWKx:''''''''''',cxKWMN0OXWMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMMMMWKd:;,,,,,,,,,,,,,,,:dKWMMMMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMMWXx:,,,,;:loddddol:;,,,,:xKWMMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMNOc;,,,;lkKNWMMMMWNKkc;,,,;cONMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMNx;,,,;oXWMMMMMMMMMMWKl;,,,;dNMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMMNk:,,:OWMMMMMMMMMMMMWk;,,:kNMMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMMMWNx:,;xNMMMMMMMMMMMMNx;,:xNWMMMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMMNKko:;,,:xNMMMMMMMMMMNx:,,;:okKNMMMMMMMMMMMMMMMM
MMMMMMMMMMMMMMMWXkdddddddxKWMMMMMMMMMXxdddddddkXWMMMMMMMMMMMMMMM\n""",
        curses.color_pair(4),
    )
    stdscr.addstr("Welcome to minFQ, the upload client for minoTour.\n")
    refresh_pad(screen, stdscr)
    try:
        validate_args(args)
        minotour_api = MinotourAPI(args.host_name, args.port_number, header)
        test_message = minotour_api.test()
        stdscr.addstr(test_message, curses.color_pair(1))
        refresh_pad(screen, stdscr)
        compatibility_message = check_server_compatibility(minotour_api, log)
        stdscr.addstr(compatibility_message, curses.color_pair(2))
        refresh_pad(screen, stdscr)
        ### List to hold folders we want to watch
        sequencing_statistics = SequencingStatistics()
        sequencing_statistics.screen_num_cols = num_cols
        sequencing_statistics.screen_num_rows = num_rows
        sequencing_statistics.minotour_url = "{}:{}".format(
            args.host_name, args.port_number
        )
        if args.list:
            list_minotour_options(log, args, minotour_api, stdscr, screen)
        # Below we perform checks if we are trying to set a job.
        if args.job is not None:
            check_job_from_client(args, log, minotour_api, parser)
        start_minknow_and_basecalled_monitoring(
            sequencing_statistics, args, header, minotour_api, stdscr, screen, log_win
        )
    except Exception as e:
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()
        curses.curs_set(1)
        curses.endwin()
        print(e)
        sys.exit(0)
